#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Boolean, Column, DateTime, ForeignKey
from sqlalchemy import Integer, MetaData, String, Table, Text


def define_tables(meta):

    services = Table(
        'services', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('host', String(length=255)),
        Column('binary', String(length=255)),
        Column('topic', String(length=255)),
        Column('report_count', Integer, nullable=False),
        Column('disabled', Boolean),
        Column('disabled_reason', String(length=255)),
        Column('modified_at', DateTime),
        Column('rpc_current_version', String(36)),
        Column('rpc_available_version', String(36)),
        mysql_engine='InnoDB'
    )

    plans = Table(
        'plans', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('name', String(length=255)),
        Column('description', String(length=255)),
        Column('provider_id', String(length=36)),
        Column('project_id', String(length=255)),
        Column('status', String(length=64)),
        Column('parameters', Text),
        mysql_engine='InnoDB'
    )

    resources = Table(
        'resources', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('plan_id', String(length=36), ForeignKey('plans.id'),
               nullable=False),
        Column('resource_id', String(length=36)),
        Column('resource_type', String(length=64)),
        Column('resource_name', String(length=255)),
        Column('resource_extra_info', Text),
        mysql_engine='InnoDB'
    )

    restores = Table(
        'restores', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean),
        Column('id', String(36), primary_key=True, nullable=False),
        Column('project_id', String(length=255)),
        Column('provider_id', String(length=36)),
        Column('checkpoint_id', String(length=36)),
        Column('restore_target', String(length=255)),
        Column('parameters', String(length=255)),
        Column('status', String(length=64)),
        Column('resources_status', Text),
        Column('resources_reason', Text),
        mysql_engine='InnoDB'
    )

    operation_logs = Table(
        'operation_logs', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean, nullable=False),
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('project_id', String(length=255), nullable=False),
        Column('scheduled_operation_id', String(length=36)),
        Column('started_at', DateTime),
        Column('ended_at', DateTime),
        Column('state', String(length=64)),
        Column('error', String(length=255)),
        Column('entries', Text),
        mysql_engine='InnoDB'
    )

    triggers = Table(
        'triggers', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean, nullable=False),
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('name', String(length=255), nullable=False),
        Column('project_id', String(length=255), nullable=False),
        Column('type', String(length=64), nullable=False),
        Column('properties', Text, nullable=False),
        mysql_engine='InnoDB'
    )

    scheduled_operations = Table(
        'scheduled_operations', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean, nullable=False),
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('name', String(length=255), nullable=False),
        Column('description', String(length=255)),
        Column('operation_type', String(length=64), nullable=False),
        Column('user_id', String(length=64), nullable=False),
        Column('project_id', String(length=255), nullable=False),
        Column('trigger_id', String(length=36), ForeignKey('triggers.id'),
               index=True, nullable=False),
        Column('operation_definition', Text, nullable=False),
        Column('enabled', Boolean, nullable=False, default=True),
        mysql_engine='InnoDB'
    )

    scheduled_operation_states = Table(
        'scheduled_operation_states',
        meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean, nullable=False),
        Column('id', Integer, primary_key=True, nullable=False,
               autoincrement=True),
        Column('operation_id', String(length=36),
               ForeignKey('scheduled_operations.id', ondelete='CASCADE'),
               index=True, unique=True, nullable=False),
        Column('service_id', Integer, ForeignKey('services.id'),
               nullable=False),
        Column('trust_id', String(length=64), nullable=False),
        Column('state', String(length=32), nullable=False),
        Column('end_time_for_run', DateTime),
        mysql_engine='InnoDB'
    )

    scheduled_operation_logs = Table(
        'scheduled_operation_logs',
        meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean, nullable=False),
        Column('id', Integer, primary_key=True, nullable=False,
               autoincrement=True),
        Column('operation_id', String(length=36),
               ForeignKey('scheduled_operations.id', ondelete='CASCADE'),
               index=True, nullable=False),
        Column('expect_start_time', DateTime),
        Column('triggered_time', DateTime),
        Column('actual_start_time', DateTime),
        Column('end_time', DateTime),
        Column('state', String(length=32), nullable=False),
        Column('extend_info', Text),
        mysql_engine='InnoDB'
    )

    checkpoint_records = Table(
        'checkpoint_records',
        meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', Boolean, nullable=False),
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('project_id', String(length=36), nullable=False),
        Column('checkpoint_id', String(length=36), nullable=False),
        Column('checkpoint_status', String(length=36), nullable=False),
        Column('provider_id', String(length=36), nullable=False),
        Column('plan_id', String(length=36), nullable=False),
        Column('operation_id', String(length=36)),
        Column('create_by', String(length=36)),
        Column('extend_info', Text),
        mysql_engine='InnoDB'
    )

    return [services,
            plans,
            resources,
            restores,
            operation_logs,
            triggers,
            scheduled_operations,
            scheduled_operation_states,
            scheduled_operation_logs,
            checkpoint_records]


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    # create all tables
    # Take care on create order for those with FK dependencies
    tables = define_tables(meta)

    for table in tables:
        table.create()

    if migrate_engine.name == "mysql":
        table_names = [t.description for t in tables]
        table_names.append("migrate_version")

        migrate_engine.execute("SET foreign_key_checks = 0")
        for table in table_names:
            migrate_engine.execute(
                "ALTER TABLE %s CONVERT TO CHARACTER SET utf8" % table)
        migrate_engine.execute("SET foreign_key_checks = 1")
        migrate_engine.execute(
            "ALTER DATABASE %s DEFAULT CHARACTER SET utf8" %
            migrate_engine.url.database)
        migrate_engine.execute("ALTER TABLE %s Engine=InnoDB" % table)


def downgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    tables = define_tables(meta)
    tables.reverse()
    for table in tables:
        table.drop()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging as log
import math
import time

from botocore.exceptions import ClientError
from karbor import exception
from karbor.i18n import _
from karbor.services.protection.bank_plugin import BankPlugin
from karbor.services.protection.bank_plugin import LeasePlugin
from karbor.services.protection import client_factory
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_service import loopingcall
from oslo_utils import uuidutils

s3_bank_plugin_opts = [
    cfg.StrOpt('bank_s3_object_bucket',
               default='karbor',
               help='The default s3 object bucket to use.'),
    cfg.StrOpt('bank_s3_lease_bucket',
               default='lease',
               help='The default s3 lease bucket to use.'),
]

LOG = logging.getLogger(__name__)
log.getLogger('botocore').setLevel(log.WARNING)

lease_opt = [cfg.IntOpt('lease_expire_window',
                        default=600,
                        help='expired_window for bank lease, in seconds'),
             cfg.IntOpt('lease_renew_window',
                        default=120,
                        help='period for bank lease, in seconds, '
                             'between bank lease client renew the lease'),
             cfg.IntOpt('lease_validity_window',
                        default=100,
                        help='validity_window for bank lease, in seconds'), ]


class S3ConnectionFailed(exception.KarborException):
    message = _("Connection to s3 failed: %(reason)s")


class S3BankPlugin(BankPlugin, LeasePlugin):
    """S3 bank plugin"""
    def __init__(self, config, context=None):
        super(S3BankPlugin, self).__init__(config)
        self._config.register_opts(s3_bank_plugin_opts,
                                   "s3_bank_plugin")
        self._config.register_opts(lease_opt,
                                   "s3_bank_plugin")
        plugin_cfg = self._config.s3_bank_plugin
        self.bank_object_bucket = plugin_cfg.bank_s3_object_bucket
        self.lease_expire_window = plugin_cfg.lease_expire_window
        self.lease_renew_window = plugin_cfg.lease_renew_window
        self.context = context
        self.lease_validity_window = plugin_cfg.lease_validity_window

        self.owner_id = uuidutils.generate_uuid()
        self.lease_expire_time = 0
        self.bank_leases_bucket = plugin_cfg.bank_s3_lease_bucket
        self._connection = None

    def _setup_connection(self):
        return client_factory.ClientFactory.create_client(
            's3',
            self.context,
            self._config
        )

    @property
    def connection(self):
        if not self._connection:
            _connection = self._setup_connection()
            # create bucket
            try:
                _connection.create_bucket(Bucket=self.bank_object_bucket)
                _connection.create_bucket(Bucket=self.bank_leases_bucket)
            except S3ConnectionFailed as err:
                LOG.error("bank plugin create bucket failed.")
                raise exception.CreateBucketrFailed(reason=err)
            self._connection = _connection

            # acquire lease
            try:
                self.acquire_lease()
            except exception.AcquireLeaseFailed:
                LOG.error("bank plugin acquire lease failed.")
                raise

            # start renew lease
            renew_lease_loop = loopingcall.FixedIntervalLoopingCall(
                self.renew_lease
            )
            renew_lease_loop.start(
                interval=self.lease_renew_window,
                initial_delay=self.lease_renew_window
            )
        return self._connection

    def get_owner_id(self):
        return self.owner_id

    def update_object(self, key, value):
        serialized = False
        try:
            if not isinstance(value, str):
                value = jsonutils.dumps(value)
                serialized = True
            self._put_object(bucket=self.bank_object_bucket,
                             obj=key,
                             contents=value,
                             headers={
                                 'x-object-meta-serialized': str(serialized)
                             })
        except S3ConnectionFailed as err:
            LOG.error("update object failed, err: %s.", err)
            raise exception.BankUpdateObjectFailed(reason=err, key=key)

    def delete_object(self, key):
        try:
            self._delete_object(bucket=self.bank_object_bucket,
                                obj=key)
        except S3ConnectionFailed as err:
            LOG.error("delete object failed, err: %s.", err)
            raise exception.BankDeleteObjectFailed(reason=err, key=key)

    def get_object(self, key):
        try:
            return self._get_object(bucket=self.bank_object_bucket,
                                    obj=key)
        except S3ConnectionFailed as err:
            LOG.error("get object failed, err: %s.", err)
            raise exception.BankGetObjectFailed(reason=err, key=key)

    def list_objects(self, prefix=None, limit=None, marker=None,
                     sort_dir=None):
        try:
            response = self._get_bucket(
                bucket=self.bank_object_bucket,
                prefix=prefix,
                limit=limit,
                marker=marker
            )
            if 'Contents' not in response:
                return []
            else:
                return [obj['Key'] for obj in response['Contents']]
        except S3ConnectionFailed as err:
            LOG.error("list objects failed, err: %s.", err)
            raise exception.BankListObjectsFailed(reason=err)

    def acquire_lease(self):
        bucket = self.bank_leases_bucket
        obj = self.owner_id
        contents = self.owner_id
        self.lease_expire_time = \
            math.floor(time.time()) + self.lease_expire_window
        headers = {'lease-expire-time': str(self.lease_expire_time)}
        try:
            self._put_object(bucket=bucket,
                             obj=obj,
                             contents=contents,
                             headers=headers)
        except S3ConnectionFailed as err:
            LOG.error("acquire lease failed, err:%s.", err)
            raise exception.AcquireLeaseFailed(reason=err)

    def renew_lease(self):
        self.acquire_lease()

    def check_lease_validity(self):
        if (self.lease_expire_time - math.floor(time.time()) >=
                self.lease_validity_window):
            return True
        else:
            self._delete_object(
                bucket=self.bank_leases_bucket,
                obj=self.owner_id
            )
            return False

    def _put_object(self, bucket, obj, contents, headers=None):
        try:
            self.connection.put_object(
                Bucket=bucket,
                Key=obj,
                Body=contents,
                Metadata=headers
            )
        except ClientError as err:
            raise S3ConnectionFailed(reason=err)

    def _get_object(self, bucket, obj):
        try:
            response = self.connection.get_object(Bucket=bucket, Key=obj)
            body = response['Body'].read()
            if response['Metadata']["x-object-meta-serialized"]\
                    .lower() == "true":
                body = jsonutils.loads(body)
            return body
        except ClientError as err:
            raise S3ConnectionFailed(reason=err)

    def _delete_object(self, bucket, obj):
        try:
            self.connection.delete_object(Bucket=bucket,
                                          Key=obj)
        except ClientError as err:
            raise S3ConnectionFailed(reason=err)

    def _get_bucket(self, bucket, prefix=None, limit=None, marker=None):
        try:
            prefix = '' if prefix is None else prefix
            marker = '' if marker is None else marker
            response = self.connection.list_objects(
                Bucket=bucket,
                Prefix=prefix,
                MaxKeys=limit,
                Marker=marker)
            return response
        except ClientError as err:
            raise S3ConnectionFailed(reason=err)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Executor which receives operations and run them.
"""

from abc import ABCMeta
from abc import abstractmethod
import six


@six.add_metaclass(ABCMeta)
class BaseExecutor(object):
    def __init__(self, operation_manager):
        self._operation_manager = operation_manager
        super(BaseExecutor, self).__init__()

    @abstractmethod
    def execute_operation(self, operation_id, triggered_time,
                          expect_start_time, window_time, **kwargs):
        """Execute an operation.

        :param operation_id: ID of operation
        :param triggered_time: time when the operation is triggered
        :param expect_start_time: expect time when to run the operation
        :param window_time: time how long to wait to run the operation after
                      expect_start_time
        """
        pass

    @abstractmethod
    def cancel_operation(self, operation_id):
        """Cancel the execution of operation.

        There is no effective for the operations which are running, but
        for operations which are in waiting, they will not be executed.

        :param operation_id: ID of operation
        """
        pass

    @abstractmethod
    def resume_operation(self, operation_id, **kwargs):
        """Resume operations.

        Get operations which are not finished from DB by operation_id,
        and execute them again.

        :param operation_id: ID of operation
        """
        pass

    @abstractmethod
    def shutdown(self):
        """Shutdown the executor"""
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of SQLAlchemy backend."""

import datetime as dt
import functools
import re
import six
import sys
import threading
import time

from oslo_config import cfg
from oslo_db import api as oslo_db_api
from oslo_db import exception as db_exc
from oslo_db import options
from oslo_db.sqlalchemy import session as db_session
from oslo_db.sqlalchemy import utils as sqlalchemyutils
from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils
from sqlalchemy import MetaData
from sqlalchemy.orm import joinedload
from sqlalchemy.schema import Table
from sqlalchemy.sql import expression
from sqlalchemy.sql.expression import literal_column
from sqlalchemy.sql import func

from karbor.db.sqlalchemy import models
from karbor import exception
from karbor.i18n import _


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

options.set_defaults(CONF, connection='sqlite:///$state_path/karbor.sqlite')

_LOCK = threading.Lock()
_FACADE = None
_GET_METHODS = {}


def _create_facade_lazily():
    global _LOCK
    with _LOCK:
        global _FACADE
        if _FACADE is None:
            _FACADE = db_session.EngineFacade(
                CONF.database.connection,
                **dict(CONF.database)
            )

        return _FACADE


def get_engine():
    facade = _create_facade_lazily()
    return facade.get_engine()


def get_session(**kwargs):
    facade = _create_facade_lazily()
    return facade.get_session(**kwargs)


def dispose_engine():
    get_engine().dispose()

_DEFAULT_QUOTA_NAME = 'default'


def get_backend():
    """The backend is this module itself."""

    return sys.modules[__name__]


def is_admin_context(context):
    """Indicates if the request context is an administrator."""
    if not context:
        LOG.warning('Use of empty request context is deprecated',
                    DeprecationWarning)
        raise Exception('die')
    return context.is_admin


def is_user_context(context):
    """Indicates if the request context is a normal user."""
    if not context:
        return False
    if context.is_admin:
        return False
    if not context.user_id or not context.project_id:
        return False
    return True


def authorize_project_context(context, project_id):
    """Ensures a request has permission to access the given project."""
    if is_user_context(context):
        if not context.project_id:
            raise exception.NotAuthorized()
        elif context.project_id != project_id:
            raise exception.NotAuthorized()


def authorize_user_context(context, user_id):
    """Ensures a request has permission to access the given user."""
    if is_user_context(context):
        if not context.user_id:
            raise exception.NotAuthorized()
        elif context.user_id != user_id:
            raise exception.NotAuthorized()


def require_admin_context(f):
    """Decorator to require admin request context.

    The first argument to the wrapped function must be the context.

    """

    def wrapper(*args, **kwargs):
        if not is_admin_context(args[0]):
            raise exception.AdminRequired()
        return f(*args, **kwargs)
    return wrapper


def require_context(f):
    """Decorator to require *any* user or admin context.

    This does no authorization for user or project access matching, see
    :py:func:`authorize_project_context` and
    :py:func:`authorize_user_context`.

    The first argument to the wrapped function must be the context.

    """

    def wrapper(*args, **kwargs):
        if not is_admin_context(args[0]) and not is_user_context(args[0]):
            raise exception.NotAuthorized()
        return f(*args, **kwargs)
    return wrapper


def require_plan_exists(f):
    """Decorator to require the specified plan to exist.

    Requires the wrapped function to use context and plan_id as
    their first two arguments.
    """
    @functools.wraps(f)
    def wrapper(context, plan_id, *args, **kwargs):
        plan_get(context, plan_id)
        return f(context, plan_id, *args, **kwargs)
    return wrapper


def _retry_on_deadlock(f):
    """Decorator to retry a DB API call if Deadlock was received."""
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        while True:
            try:
                return f(*args, **kwargs)
            except db_exc.DBDeadlock:
                LOG.warning("Deadlock detected when running '%(func_name)s':"
                            " Retrying...", dict(func_name=f.__name__))
                # Retry!
                time.sleep(0.5)
                continue
    functools.update_wrapper(wrapped, f)
    return wrapped


def model_query(context, *args, **kwargs):
    """Query helper that accounts for context's `read_deleted` field.

    :param context: context to query under
    :param session: if present, the session to use
    :param read_deleted: if present, overrides context's read_deleted field.
    :param project_only: if present and context is user-type, then restrict
            query to match the context's project_id.
    """
    session = kwargs.get('session') or get_session()
    read_deleted = kwargs.get('read_deleted') or context.read_deleted
    project_only = kwargs.get('project_only')

    query = session.query(*args)
    if read_deleted == 'no':
        query = query.filter_by(deleted=False)
    elif read_deleted == 'yes':
        pass  # omit the filter to include deleted and active
    elif read_deleted == 'only':
        query = query.filter_by(deleted=True)
    else:
        raise Exception(
            _("Unrecognized read_deleted value '%s'") % read_deleted)

    if project_only and is_user_context(context):
        query = query.filter_by(project_id=context.project_id)

    return query


@require_admin_context
def service_destroy(context, service_id):
    session = get_session()
    with session.begin():
        service_ref = _service_get(context, service_id, session=session)
        service_ref.delete(session=session)


@require_admin_context
def _service_get(context, service_id, session=None):
    result = model_query(
        context,
        models.Service,
        session=session
    ).filter_by(id=service_id).first()
    if not result:
        raise exception.ServiceNotFound(service_id=service_id)

    return result


@require_admin_context
def service_get(context, service_id):
    return _service_get(context, service_id)


@require_admin_context
def service_get_all(context, disabled=None):
    query = model_query(context, models.Service)

    if disabled is not None:
        query = query.filter_by(disabled=disabled)

    return query.all()


@require_admin_context
def service_get_all_by_topic(context, topic, disabled=None):
    query = model_query(
        context,
        models.Service,
        read_deleted="no"
    ).filter_by(topic=topic)

    if disabled is not None:
        query = query.filter_by(disabled=disabled)

    return query.all()


@require_admin_context
def service_get_by_host_and_topic(context, host, topic):
    result = model_query(
        context,
        models.Service,
        read_deleted="no"
    ).filter_by(
        disabled=False
    ).filter_by(
        host=host
    ).filter_by(
        topic=topic
    ).first()
    if not result:
        raise exception.ServiceNotFound(service_id=None)
    return result


@require_admin_context
def _service_get_all_topic_subquery(context, session, topic, subq, label):
    sort_value = getattr(subq.c, label)
    return model_query(
        context,
        models.Service,
        func.coalesce(sort_value, 0),
        session=session,
        read_deleted="no"
    ).filter_by(
        topic=topic
    ).filter_by(
        disabled=False
    ).outerjoin(
        (subq, models.Service.host == subq.c.host)
    ).order_by(
        sort_value
    ).all()


@require_admin_context
def service_get_by_args(context, host, binary):
    results = model_query(
        context,
        models.Service
    ).filter_by(
        host=host
    ).filter_by(
        binary=binary
    ).all()

    for result in results:
        if host == result['host']:
            return result

    raise exception.HostBinaryNotFound(host=host, binary=binary)


@require_admin_context
def service_create(context, values):
    service_ref = models.Service()
    service_ref.update(values)
    if not CONF.enable_new_services:
        service_ref.disabled = True

    session = get_session()
    with session.begin():
        service_ref.save(session)
        return service_ref


@require_admin_context
def service_update(context, service_id, values):
    session = get_session()
    with session.begin():
        service_ref = _service_get(context, service_id, session=session)
        if ('disabled' in values):
            service_ref['modified_at'] = timeutils.utcnow()
            service_ref['updated_at'] = literal_column('updated_at')
        service_ref.update(values)
        return service_ref


def _get_get_method(model):
    # General conversion
    # Convert camel cased model name to snake format
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', model.__name__)
    # Get method must be snake formatted model name concatenated with _get
    method_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower() + '_get'
    return globals().get(method_name)


@require_context
def get_by_id(context, model, id, *args, **kwargs):
    # Add get method to cache dictionary if it's not already there
    if not _GET_METHODS.get(model):
        _GET_METHODS[model] = _get_get_method(model)

    return _GET_METHODS[model](context, id, *args, **kwargs)


###################


def trigger_get(context, id):
    return _trigger_get(context, id)


def _trigger_get(context, id, session=None):
    result = model_query(context, models.Trigger,
                         session=session).filter_by(id=id)
    result = result.first()

    if not result:
        raise exception.TriggerNotFound(id=id)

    return result


def trigger_create(context, values):
    if not values.get('id'):
        values['id'] = uuidutils.generate_uuid()

    trigger_ref = models.Trigger()
    trigger_ref.update(values)
    trigger_ref.save(get_session())
    return trigger_ref


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
def trigger_update(context, id, values):
    """Update the Trigger record with the most recent data."""

    session = get_session()
    with session.begin():
        trigger_ref = _trigger_get(context, id, session=session)
        trigger_ref.update(values)
        trigger_ref.save(session)
    return trigger_ref


def trigger_delete(context, id):
    """Delete a Trigger record."""

    session = get_session()
    with session.begin():
        trigger_ref = _trigger_get(context, id, session=session)
        trigger_ref.delete(session=session)


def _trigger_list_query(context, session, **kwargs):
    return model_query(context, models.Trigger, session=session)


def _trigger_list_process_filters(query, filters):
    exact_match_filter_names = ['project_id', 'type']
    query = _list_common_process_exact_filter(models.Trigger, query, filters,
                                              exact_match_filter_names)

    regex_match_filter_names = ['name', 'properties']
    query = _list_common_process_regex_filter(models.Trigger, query, filters,
                                              regex_match_filter_names)

    return query


def trigger_get_all_by_filters_sort(context, filters, limit=None, marker=None,
                                    sort_keys=None, sort_dirs=None):
    session = get_session()
    with session.begin():
        query = _generate_paginate_query(context, session, marker, limit,
                                         sort_keys, sort_dirs, filters,
                                         paginate_type=models.Trigger,
                                         use_model=True)

        return query.all() if query else []


###################


def scheduled_operation_get(context, id, columns_to_join=[]):
    return _scheduled_operation_get(context, id,
                                    columns_to_join=columns_to_join)


def _scheduled_operation_get(context, id, columns_to_join=[], session=None):
    query = model_query(context, models.ScheduledOperation,
                        session=session).filter_by(id=id)

    if columns_to_join and 'trigger' in columns_to_join:
        query = query.options(joinedload('trigger'))

    result = query.first()
    if not result:
        raise exception.ScheduledOperationNotFound(id=id)

    return result


def scheduled_operation_create(context, values):
    if not values.get('id'):
        values['id'] = uuidutils.generate_uuid()

    operation_ref = models.ScheduledOperation()
    operation_ref.update(values)
    operation_ref.save(get_session())
    return operation_ref


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
def scheduled_operation_update(context, id, values):
    """Update the ScheduledOperation record with the most recent data."""

    session = get_session()
    with session.begin():
        operation_ref = _scheduled_operation_get(context, id,
                                                 session=session)
        operation_ref.update(values)
        operation_ref.save(session)
    return operation_ref


def scheduled_operation_delete(context, id):
    """Delete a ScheduledOperation record."""

    session = get_session()
    with session.begin():
        operation_ref = _scheduled_operation_get(context, id,
                                                 session=session)
        session.delete(operation_ref)
        session.flush()


def _scheduled_operation_list_query(context, session, **kwargs):
    return model_query(context, models.ScheduledOperation, session=session)


def _scheduled_operation_list_process_filters(query, filters):
    exact_match_filter_names = ['project_id', 'operation_type', 'trigger_id']
    query = _list_common_process_exact_filter(
        models.ScheduledOperation, query, filters,
        exact_match_filter_names)

    regex_match_filter_names = ['name', 'operation_definition']
    query = _list_common_process_regex_filter(
        models.ScheduledOperation, query, filters,
        regex_match_filter_names)

    return query


def scheduled_operation_get_all_by_filters_sort(
        context, filters, limit=None, marker=None,
        sort_keys=None, sort_dirs=None):

    session = get_session()
    with session.begin():
        query = _generate_paginate_query(
            context, session, marker, limit,
            sort_keys, sort_dirs, filters,
            paginate_type=models.ScheduledOperation,
            use_model=True)

        return query.all() if query else []


###################


def scheduled_operation_state_get(context, operation_id, columns_to_join=[]):
    return _scheduled_operation_state_get(context, operation_id,
                                          columns_to_join=columns_to_join)


def _scheduled_operation_state_get(context, operation_id,
                                   columns_to_join=[], session=None):
    query = model_query(context, models.ScheduledOperationState,
                        session=session).filter_by(operation_id=operation_id)

    if columns_to_join and 'operation' in columns_to_join:
        query = query.options(joinedload('operation'))

    result = query.first()
    if not result:
        raise exception.ScheduledOperationStateNotFound(op_id=operation_id)
    return result


def scheduled_operation_state_create(context, values):
    state_ref = models.ScheduledOperationState()
    state_ref.update(values)
    state_ref.save(get_session())
    return state_ref


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
def scheduled_operation_state_update(context, operation_id, values):
    """Update the ScheduledOperationState record with the most recent data."""

    session = get_session()
    with session.begin():
        state_ref = _scheduled_operation_state_get(context, operation_id,
                                                   session=session)
        state_ref.update(values)
        state_ref.save(session)
    return state_ref


def scheduled_operation_state_delete(context, operation_id):
    """Delete a ScheduledOperationState record."""

    session = get_session()
    with session.begin():
        state_ref = _scheduled_operation_state_get(context, operation_id,
                                                   session=session)
        session.delete(state_ref)
        session.flush()


def _scheduled_operation_state_list_query(context, session, **kwargs):
    query = model_query(context, models.ScheduledOperationState,
                        session=session)

    valid_columns = ['operation']
    columns_to_join = kwargs.get('columns_to_join', [])
    for column in columns_to_join:
        if column in valid_columns:
            query = query.options(joinedload(column))

    return query


def _scheduled_operation_state_list_process_filters(query, filters):
    exact_match_filter_names = ['service_id', 'state']
    query = _list_common_process_exact_filter(
        models.ScheduledOperationState, query, filters,
        exact_match_filter_names)

    return query


def scheduled_operation_state_get_all_by_filters_sort(
        context, filters, limit=None, marker=None,
        sort_keys=None, sort_dirs=None, columns_to_join=[]):

    session = get_session()
    with session.begin():
        query = _generate_paginate_query(
            context, session, marker, limit,
            sort_keys, sort_dirs, filters,
            paginate_type=models.ScheduledOperationState,
            use_model=True, columns_to_join=columns_to_join)

        return query.all() if query else []


###################


def scheduled_operation_log_get(context, log_id):
    return _scheduled_operation_log_get(context, log_id)


def _scheduled_operation_log_get(context, log_id, session=None):
    result = model_query(context, models.ScheduledOperationLog,
                         session=session).filter_by(id=log_id).first()

    if not result:
        raise exception.ScheduledOperationLogNotFound(log_id=log_id)

    return result


def scheduled_operation_log_create(context, values):
    log_ref = models.ScheduledOperationLog()
    log_ref.update(values)
    log_ref.save(get_session())
    return log_ref


@oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
def scheduled_operation_log_update(context, log_id, values):
    """Update the ScheduledOperationLog record with the most recent data."""

    session = get_session()
    with session.begin():
        log_ref = _scheduled_operation_log_get(context, log_id,
                                               session=session)
        log_ref.update(values)
        log_ref.save(session)
    return log_ref


def scheduled_operation_log_delete(context, log_id):
    """Delete a ScheduledOperationLog record."""

    session = get_session()
    with session.begin():
        log_ref = _scheduled_operation_log_get(context, log_id,
                                               session=session)
        session.delete(log_ref)
        session.flush()


def scheduled_operation_log_delete_oldest(context, operation_id,
                                          retained_num, excepted_states):
    table = models.ScheduledOperationLog
    session = get_session()
    with session.begin():
        result = model_query(context, table, session=session).filter_by(
            operation_id=operation_id).order_by(
            expression.desc(table.created_at)).limit(retained_num).all()

        if not result or len(result) < retained_num:
            return
        oldest_create_time = result[-1]['created_at']

        if excepted_states and isinstance(excepted_states, list):
            filters = expression.and_(
                table.operation_id == operation_id,
                table.created_at < oldest_create_time,
                table.state.notin_(excepted_states))
        else:
            filters = expression.and_(
                table.operation_id == operation_id,
                table.created_at < oldest_create_time)

        model_query(context, table, session=session).filter(
            filters).delete(synchronize_session=False)


def _scheduled_operation_log_list_query(context, session, **kwargs):
    query = model_query(context, models.ScheduledOperationLog,
                        session=session)
    return query


def _scheduled_operation_log_list_process_filters(query, filters):
    exact_match_filter_names = ['operation_id', 'state']
    query = _list_common_process_exact_filter(
        models.ScheduledOperationLog, query, filters,
        exact_match_filter_names)

    return query


def scheduled_operation_log_get_all_by_filters_sort(
        context, filters, limit=None, marker=None,
        sort_keys=None, sort_dirs=None):

    session = get_session()
    with session.begin():
        query = _generate_paginate_query(
            context, session, marker, limit,
            sort_keys, sort_dirs, filters,
            paginate_type=models.ScheduledOperationLog,
            use_model=True)

        return query.all() if query else []


###################


def _resource_refs(resource_list, meta_class):
    resource_refs = []
    if resource_list:
        for resource in resource_list:
            resource_ref = meta_class()
            resource_ref['resource_id'] = resource['id']
            resource_ref['resource_type'] = resource['type']
            resource_ref['resource_name'] = resource['name']
            resource_ref['resource_extra_info'] = resource.get(
                'extra_info', None)
            resource_refs.append(resource_ref)
    return resource_refs


@require_context
def _plan_get_query(context, session=None, project_only=False,
                    joined_load=True):
    """Get the query to retrieve the plan.

    :param context: the context used to run the method _plan_get_query
    :param session: the session to use
    :param project_only: the boolean used to decide whether to query the
                         plan in the current project or all projects
    :param joined_load: the boolean used to decide whether the query loads
                        the other models, which join the plan model in
                        the database.
    :returns: updated query or None
    """
    query = model_query(context, models.Plan, session=session,
                        project_only=project_only)
    if joined_load:
        query = query.options(joinedload('resources'))
    return query


def _plan_resources_get_query(context, plan_id, model, session=None):
    return model_query(
        context,
        model,
        session=session,
        read_deleted="no"
    ).filter_by(plan_id=plan_id)


@require_context
def _resource_create(context, values):
    resource_ref = models.Resource()
    resource_ref.update(values)
    session = get_session()
    with session.begin():
        resource_ref.save(session)
        return resource_ref


@require_context
def _plan_resources_update(context, plan_id, resources, session=None):
    session = session or get_session()
    now = timeutils.utcnow()
    with session.begin():
        model_query(
            context,
            models.Resource,
            session=session
        ).filter_by(
            plan_id=plan_id
        ).update({
            'deleted': True,
            'deleted_at': now,
            'updated_at': literal_column('updated_at')
        })
    resources_list = []
    for resource in resources:
        resource['plan_id'] = plan_id
        resource['resource_id'] = resource.pop('id')
        resource['resource_type'] = resource.pop('type')
        resource['resource_name'] = resource.pop('name')
        resource['resource_extra_info'] = resource.pop(
            'extra_info', None)
        resource_ref = _resource_create(context, resource)
        resources_list.append(resource_ref)

    return resources_list


@require_context
def _plan_get(context, plan_id, session=None, joined_load=True):
    result = _plan_get_query(context, session=session, project_only=True,
                             joined_load=joined_load)
    result = result.filter_by(id=plan_id).first()

    if not result:
        raise exception.PlanNotFound(plan_id=plan_id)

    return result


@require_context
def plan_create(context, values):
    values['resources'] = _resource_refs(values.get('resources'),
                                         models.Resource)

    plan_ref = models.Plan()
    if not values.get('id'):
        values['id'] = uuidutils.generate_uuid()
    plan_ref.update(values)

    session = get_session()
    with session.begin():
        session.add(plan_ref)

    return _plan_get(context, values['id'], session=session)


@require_context
def plan_get(context, plan_id):
    return _plan_get(context, plan_id)


@require_admin_context
@_retry_on_deadlock
def plan_destroy(context, plan_id):
    session = get_session()
    now = timeutils.utcnow()
    with session.begin():
        model_query(
            context,
            models.Plan,
            session=session
        ).filter_by(
            id=plan_id
        ).update({
            'deleted': True,
            'deleted_at': now,
            'updated_at': literal_column('updated_at')
        })
        model_query(
            context,
            models.Resource,
            session=session
        ).filter_by(
            plan_id=plan_id
        ).update({
            'deleted': True,
            'deleted_at': now,
            'updated_at': literal_column('updated_at')
        })


@require_context
@require_plan_exists
def plan_update(context, plan_id, values):
    session = get_session()
    with session.begin():
        resources = values.get('resources')
        if resources is not None:
            _plan_resources_update(context,
                                   plan_id,
                                   values.pop('resources'),
                                   session=session)

        plan_ref = _plan_get(context, plan_id, session=session)
        plan_ref.update(values)

        return plan_ref


@require_context
@require_plan_exists
@_retry_on_deadlock
def plan_resources_update(context, plan_id, resources):
    return _plan_resources_update(context,
                                  plan_id,
                                  resources)


@require_admin_context
def plan_get_all(context, marker, limit, sort_keys=None, sort_dirs=None,
                 filters=None, offset=None):
    """Retrieves all plans.

    If no sort parameters are specified then the returned plans are sorted
    first by the 'created_at' key and then by the 'id' key in descending
    order.

    :param context: context to query under
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param filters: dictionary of filters; values that are in lists, tuples,
                    or sets cause an 'IN' operation, while exact matching
                    is used for other values, see _process_plan_filters
                    function for more information
    :returns: list of matching plans
    """
    session = get_session()
    with session.begin():
        # Generate the query
        query = _generate_paginate_query(context, session, marker, limit,
                                         sort_keys, sort_dirs, filters, offset)
        # No plans would match, return empty list
        if query is None:
            return []
        return query.all()


@require_context
def plan_get_all_by_project(context, project_id, marker, limit,
                            sort_keys=None, sort_dirs=None, filters=None,
                            offset=None):
    """Retrieves all plans in a project.

    If no sort parameters are specified then the returned plans are sorted
    first by the 'created_at' key and then by the 'id' key in descending
    order.

    :param context: context to query under
    :param project_id: project for all plans being retrieved
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param filters: dictionary of filters; values that are in lists, tuples,
                    or sets cause an 'IN' operation, while exact matching
                    is used for other values, see _process_plan_filters
                    function for more information
    :returns: list of matching plans
    """
    session = get_session()
    with session.begin():
        authorize_project_context(context, project_id)
        # Add in the project filter without modifying the given filters
        filters = filters.copy() if filters else {}
        filters['project_id'] = project_id
        # Generate the query
        query = _generate_paginate_query(context, session, marker, limit,
                                         sort_keys, sort_dirs, filters, offset)
        # No plans would match, return empty list
        if query is None:
            return []
        return query.all()


def _process_plan_filters(query, filters):
    exact_match_filter_names = ['project_id', 'status']
    query = _list_common_process_exact_filter(models.Plan, query, filters,
                                              exact_match_filter_names)

    regex_match_filter_names = ['name', 'description']
    query = _list_common_process_regex_filter(models.Plan, query, filters,
                                              regex_match_filter_names)

    return query


###############################


@require_context
def restore_create(context, values):
    restore_ref = models.Restore()
    if not values.get('id'):
        values['id'] = uuidutils.generate_uuid()
    restore_ref.update(values)

    session = get_session()
    with session.begin():
        restore_ref.save(session)
        return restore_ref


@require_context
def restore_get(context, restore_id):
    return _restore_get(context, restore_id)


@require_context
def _restore_get(context, restore_id, session=None):
    result = model_query(
        context,
        models.Restore,
        session=session
    ).filter_by(
        id=restore_id
    ).first()
    if not result:
        raise exception.RestoreNotFound(restore_id=restore_id)

    return result


@require_context
def restore_update(context, restore_id, values):
    session = get_session()
    with session.begin():
        restore_ref = _restore_get(context, restore_id, session=session)
        restore_ref.update(values)
        return restore_ref


@require_context
@_retry_on_deadlock
def restore_destroy(context, restore_id):
    session = get_session()
    with session.begin():
        restore_ref = _restore_get(context, restore_id, session=session)
        restore_ref.delete(session=session)


def is_valid_model_filters(model, filters):
    """Return True if filter values exist on the model

    :param model: a karbor model
    :param filters: dictionary of filters
    """
    for key in filters.keys():
        try:
            getattr(model, key)
        except AttributeError:
            LOG.debug("'%s' filter key is not valid.", key)
            return False
    return True


def _restore_get_query(context, session=None, project_only=False):
    return model_query(context, models.Restore, session=session,
                       project_only=project_only)


@require_admin_context
def restore_get_all(context, marker, limit, sort_keys=None, sort_dirs=None,
                    filters=None, offset=None):
    """Retrieves all restores.

    If no sort parameters are specified then the returned plans are sorted
    first by the 'created_at' key and then by the 'id' key in descending
    order.

    :param context: context to query under
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param filters: dictionary of filters; values that are in lists, tuples,
                    or sets cause an 'IN' operation, while exact matching
                    is used for other values, see _process_plan_filters
                    function for more information
    :returns: list of matching restores
    """
    if filters and not is_valid_model_filters(models.Restore, filters):
        return []

    session = get_session()
    with session.begin():
        # Generate the query
        query = _generate_paginate_query(context, session, marker, limit,
                                         sort_keys, sort_dirs, filters,
                                         offset, models.Restore)
        # No restores would match, return empty list
        if query is None:
            return []
        return query.all()


@require_context
def restore_get_all_by_project(context, project_id, marker, limit,
                               sort_keys=None, sort_dirs=None, filters=None,
                               offset=None):
    """Retrieves all restores in a project.

    If no sort parameters are specified then the returned plans are sorted
    first by the 'created_at' key and then by the 'id' key in descending
    order.

    :param context: context to query under
    :param project_id: project for all plans being retrieved
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param filters: dictionary of filters; values that are in lists, tuples,
                    or sets cause an 'IN' operation, while exact matching
                    is used for other values, see _process_plan_filters
                    function for more information
    :returns: list of matching restores
    """
    if filters and not is_valid_model_filters(models.Restore, filters):
        return []

    session = get_session()
    with session.begin():
        authorize_project_context(context, project_id)
        # Add in the project filter without modifying the given filters
        filters = filters.copy() if filters else {}
        filters['project_id'] = project_id
        # Generate the query
        query = _generate_paginate_query(context, session, marker, limit,
                                         sort_keys, sort_dirs, filters,
                                         offset, models.Restore)
        # No plans would match, return empty list
        if query is None:
            return []
        return query.all()


def _process_restore_filters(query, filters):
    if filters:
        # Ensure that filters' keys exist on the model
        if not is_valid_model_filters(models.Restore, filters):
            return None
        query = query.filter_by(**filters)
    return query


###############################


@require_context
def operation_log_create(context, values):
    operation_log_ref = models.OperationLog()
    if not values.get('id'):
        values['id'] = uuidutils.generate_uuid()
    operation_log_ref.update(values)

    session = get_session()
    with session.begin():
        operation_log_ref.save(session)
        return operation_log_ref


@require_context
def operation_log_get(context, operation_log_id):
    return _operation_log_get(context, operation_log_id)


@require_context
def _operation_log_get(context, operation_log_id, session=None):
    result = model_query(
        context,
        models.OperationLog,
        session=session
    ).filter_by(
        id=operation_log_id
    ).first()
    if not result:
        raise exception.OperationLogNotFound(restore_id=operation_log_id)

    return result


@require_context
def operation_log_update(context, operation_log_id, values):
    session = get_session()
    with session.begin():
        operation_log_ref = _operation_log_get(context, operation_log_id,
                                               session=session)
        operation_log_ref.update(values)
        return operation_log_ref


@require_context
@_retry_on_deadlock
def operation_log_destroy(context, operation_log_id):
    session = get_session()
    with session.begin():
        operation_log_ref = _operation_log_get(context, operation_log_id,
                                               session=session)
        operation_log_ref.delete(session=session)


def _operation_log_get_query(context, session=None, project_only=False):
    return model_query(context, models.OperationLog, session=session,
                       project_only=project_only)


@require_admin_context
def operation_log_get_all(context, marker, limit, sort_keys=None,
                          sort_dirs=None,
                          filters=None, offset=None):
    """Retrieves all operation logs.

    If no sort parameters are specified then the returned plans are sorted
    first by the 'created_at' key and then by the 'id' key in descending
    order.

    :param context: context to query under
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param filters: dictionary of filters; values that are in lists, tuples,
                    or sets cause an 'IN' operation, while exact matching
                    is used for other values, see _process_plan_filters
                    function for more information
    :returns: list of matching operation logs
    """
    if filters and not is_valid_model_filters(models.OperationLog, filters):
        return []

    session = get_session()
    with session.begin():
        # Generate the query
        query = _generate_paginate_query(context, session, marker, limit,
                                         sort_keys, sort_dirs, filters,
                                         offset, models.OperationLog)
        # No restores would match, return empty list
        if query is None:
            return []
        return query.all()


@require_context
def operation_log_get_all_by_project(context, project_id, marker, limit,
                                     sort_keys=None, sort_dirs=None,
                                     filters=None,
                                     offset=None):
    """Retrieves all operation logs in a project.

    If no sort parameters are specified then the returned plans are sorted
    first by the 'created_at' key and then by the 'id' key in descending
    order.

    :param context: context to query under
    :param project_id: project for all plans being retrieved
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param filters: dictionary of filters; values that are in lists, tuples,
                    or sets cause an 'IN' operation, while exact matching
                    is used for other values, see _process_plan_filters
                    function for more information
    :returns: list of matching restores
    """
    if filters and not is_valid_model_filters(models.OperationLog, filters):
        return []

    session = get_session()
    with session.begin():
        authorize_project_context(context, project_id)
        # Add in the project filter without modifying the given filters
        filters = filters.copy() if filters else {}
        filters['project_id'] = project_id
        # Generate the query
        query = _generate_paginate_query(context, session, marker, limit,
                                         sort_keys, sort_dirs, filters,
                                         offset, models.OperationLog)
        # No plans would match, return empty list
        if query is None:
            return []
        return query.all()


def _process_operation_log_filters(query, filters):
    if filters:
        # Ensure that filters' keys exist on the model
        if not is_valid_model_filters(models.OperationLog, filters):
            return None
        query = query.filter_by(**filters)
    return query
###############################


@require_context
def checkpoint_record_create(context, values):
    checkpoint_record_ref = models.CheckpointRecord()
    if not values.get('id'):
        values['id'] = uuidutils.generate_uuid()
    checkpoint_record_ref.update(values)

    session = get_session()
    with session.begin():
        checkpoint_record_ref.save(session)
        return checkpoint_record_ref


@require_context
def checkpoint_record_get(context, checkpoint_record_id):
    return _checkpoint_record_get(context, checkpoint_record_id)


@require_context
def _checkpoint_record_get(context, checkpoint_record_id, session=None):
    result = model_query(
        context,
        models.CheckpointRecord,
        session=session).filter_by(
            id=checkpoint_record_id).first()
    if not result:
        raise exception.CheckpointRecordNotFound(id=checkpoint_record_id)

    return result


@require_context
def checkpoint_record_update(context, checkpoint_record_id, values):
    session = get_session()
    with session.begin():
        checkpoint_record_ref = _checkpoint_record_get(context,
                                                       checkpoint_record_id,
                                                       session=session)
        checkpoint_record_ref.update(values)
        return checkpoint_record_ref


@require_context
@_retry_on_deadlock
def checkpoint_record_destroy(context, checkpoint_record_id):
    session = get_session()
    with session.begin():
        checkpoint_record_ref = _checkpoint_record_get(context,
                                                       checkpoint_record_id,
                                                       session=session)
        checkpoint_record_ref.delete(session=session)


def _checkpoint_record_list_query(context, session, **kwargs):
    return model_query(context, models.CheckpointRecord, session=session)


def _checkpoint_record_list_process_filters(query, filters):
    exact_match_filter_names = ['project_id', 'id',
                                'checkpoint_id', 'checkpoint_status',
                                'plan_id', 'provider_id', 'operation_id']
    query = _list_common_process_exact_filter(
        models.CheckpointRecord, query, filters,
        exact_match_filter_names)

    regex_match_filter_names = ['create_by']
    query = _list_common_process_regex_filter(
        models.CheckpointRecord, query, filters,
        regex_match_filter_names)

    return query


def checkpoint_record_get_all_by_filters_sort(
        context, filters, limit=None, marker=None,
        sort_keys=None, sort_dirs=None):

    session = get_session()
    with session.begin():
        query = _generate_paginate_query(
            context, session, marker, limit,
            sort_keys, sort_dirs, filters,
            paginate_type=models.CheckpointRecord,
            use_model=True)

        return query.all() if query else []
###############################


@require_context
def _list_common_get_query(context, model, session=None):
    return model_query(context, model, session=session)


def _list_common_process_exact_filter(model, query, filters, legal_keys):
    """Applies exact match filtering to a query.

    :param model: model to apply filters to
    :param query: query to apply filters to
    :param filters: dictionary of filters; values that are lists,
                    tuples, sets, or frozensets cause an 'IN' test to
                    be performed, while exact matching ('==' operator)
                    is used for other values
    :param legal_keys: list of keys to apply exact filtering to
    :returns: the updated query.
    """

    filter_dict = {}
    for key in legal_keys:
        if key not in filters:
            continue

        value = filters.get(key)
        if isinstance(value, (list, tuple, set, frozenset)):
            if not value:
                return None  # empty IN-predicate; short circuit
            # Looking for values in a list; apply to query directly
            column_attr = getattr(model, key)
            query = query.filter(column_attr.in_(value))
        else:
            # OK, simple exact match; save for later
            filter_dict[key] = value

    # Apply simple exact matches
    if filter_dict:
        query = query.filter_by(**filter_dict)

    return query


def _list_common_process_regex_filter(model, query, filters, legal_keys):
    """Applies regular expression filtering to a query.

    :param model: model to apply filters to
    :param query: query to apply filters to
    :param filters: dictionary of filters with regex values
    :param legal_keys: list of keys to apply regex filtering to
    :returns: the updated query.
    """

    def _get_regexp_op_for_connection(db_connection):
        db_string = db_connection.split(':')[0].split('+')[0]
        regexp_op_map = {
            'postgresql': '~',
            'mysql': 'REGEXP',
            'sqlite': 'REGEXP'
        }
        return regexp_op_map.get(db_string, 'LIKE')

    db_regexp_op = _get_regexp_op_for_connection(CONF.database.connection)
    for key in legal_keys:
        if key not in filters:
            continue

        value = filters[key]
        if not isinstance(value, six.string_types):
            continue

        column_attr = getattr(model, key)
        if db_regexp_op == 'LIKE':
            query = query.filter(column_attr.op(db_regexp_op)(
                u'%' + value + u'%'))
        else:
            query = query.filter(column_attr.op(db_regexp_op)(
                value))
    return query


PAGINATION_HELPERS = {
    models.Plan: (_plan_get_query, _process_plan_filters, _plan_get),
    models.Restore: (_restore_get_query, _process_restore_filters,
                     _restore_get),
    models.Trigger: (_trigger_list_query, _trigger_list_process_filters,
                     _trigger_get),
    models.ScheduledOperation: (_scheduled_operation_list_query,
                                _scheduled_operation_list_process_filters,
                                _scheduled_operation_get),

    models.ScheduledOperationState: (
        _scheduled_operation_state_list_query,
        _scheduled_operation_state_list_process_filters,
        _scheduled_operation_state_get),

    models.OperationLog: (_operation_log_get_query,
                          _process_operation_log_filters,
                          _operation_log_get),

    models.ScheduledOperationLog: (
        _scheduled_operation_log_list_query,
        _scheduled_operation_log_list_process_filters,
        _scheduled_operation_log_get),

    models.CheckpointRecord: (
        _checkpoint_record_list_query,
        _checkpoint_record_list_process_filters,
        _checkpoint_record_get),
}


###############################


def _generate_paginate_query(context, session, marker, limit, sort_keys,
                             sort_dirs, filters, offset=None,
                             paginate_type=models.Plan, use_model=False,
                             **kwargs):
    """Generate the query to include the filters and the paginate options.

    Returns a query with sorting / pagination criteria added or None
    if the given filters will not yield any results.

    :param context: context to query under
    :param session: the session to use
    :param marker: the last item of the previous page; we returns the next
                    results after this value.
    :param limit: maximum number of items to return
    :param sort_keys: list of attributes by which results should be sorted,
                      paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                      paired with corresponding item in sort_keys
    :param filters: dictionary of filters; values that are in lists, tuples,
                    or sets cause an 'IN' operation, while exact matching
                    is used for other values, see _process_plan_filters
                    function for more information
    :param offset: number of items to skip
    :param paginate_type: type of pagination to generate
    :returns: updated query or None
    """
    get_query, process_filters, get = PAGINATION_HELPERS[paginate_type]

    sort_keys, sort_dirs = process_sort_params(sort_keys,
                                               sort_dirs,
                                               default_dir='desc')
    if use_model:
        query = get_query(context, session=session, **kwargs)
    else:
        query = get_query(context, session=session)

    if filters:
        query = process_filters(query, filters)
        if query is None:
            return None

    marker_object = None
    if marker is not None:
        marker_object = get(context, marker, session=session)

    query = sqlalchemyutils.paginate_query(query, paginate_type, limit,
                                           sort_keys,
                                           marker=marker_object,
                                           sort_dirs=sort_dirs)
    if offset:
        query = query.offset(offset)
    return query


def process_sort_params(sort_keys, sort_dirs, default_keys=None,
                        default_dir='asc'):
    """Process the sort parameters to include default keys.

    Creates a list of sort keys and a list of sort directions. Adds the default
    keys to the end of the list if they are not already included.

    When adding the default keys to the sort keys list, the associated
    direction is:
    1) The first element in the 'sort_dirs' list (if specified), else
    2) 'default_dir' value (Note that 'asc' is the default value since this is
    the default in sqlalchemy.utils.paginate_query)

    :param sort_keys: List of sort keys to include in the processed list
    :param sort_dirs: List of sort directions to include in the processed list
    :param default_keys: List of sort keys that need to be included in the
                         processed list, they are added at the end of the list
                         if not already specified.
    :param default_dir: Sort direction associated with each of the default
                        keys that are not supplied, used when they are added
                        to the processed list
    :returns: list of sort keys, list of sort directions
    :raise exception.InvalidInput: If more sort directions than sort keys
                                   are specified or if an invalid sort
                                   direction is specified
    """
    if default_keys is None:
        default_keys = ['created_at', 'id']

    # Determine direction to use for when adding default keys
    if sort_dirs and len(sort_dirs):
        default_dir_value = sort_dirs[0]
    else:
        default_dir_value = default_dir

    # Create list of keys (do not modify the input list)
    if sort_keys:
        result_keys = list(sort_keys)
    else:
        result_keys = []

    # If a list of directions is not provided, use the default sort direction
    # for all provided keys.
    if sort_dirs:
        result_dirs = []
        # Verify sort direction
        for sort_dir in sort_dirs:
            if sort_dir not in ('asc', 'desc'):
                msg = _("Unknown sort direction, must be 'desc' or 'asc'.")
                raise exception.InvalidInput(reason=msg)
            result_dirs.append(sort_dir)
    else:
        result_dirs = [default_dir_value for _sort_key in result_keys]

    # Ensure that the key and direction length match
    while len(result_dirs) < len(result_keys):
        result_dirs.append(default_dir_value)
    # Unless more direction are specified, which is an error
    if len(result_dirs) > len(result_keys):
        msg = _("Sort direction array size exceeds sort key array size.")
        raise exception.InvalidInput(reason=msg)

    # Ensure defaults are included
    for key in default_keys:
        if key not in result_keys:
            result_keys.append(key)
            result_dirs.append(default_dir_value)

    return result_keys, result_dirs


@require_admin_context
def purge_deleted_rows(context, age_in_days):
    """Purge deleted rows older than age from karbor tables."""
    try:
        age_in_days = int(age_in_days)
    except ValueError:
        msg = _('Invalid valude for age, %(age)s')
        LOG.exception(msg, {'age': age_in_days})
        raise exception.InvalidParameterValue(msg % {'age': age_in_days})
    if age_in_days <= 0:
        msg = _('Must supply a positive value for age')
        LOG.exception(msg)
        raise exception.InvalidParameterValue(msg)

    engine = get_engine()
    session = get_session()
    metadata = MetaData()
    metadata.bind = engine
    tables = []

    for model_class in models.__dict__.values():
        if hasattr(model_class, "__tablename__") and hasattr(
                model_class, "deleted"):
            tables.append(model_class.__tablename__)

    # Reorder the list so the tables are last to avoid ForeignKey constraints
    # get rid of FK constraints
    for tbl in ('plans', 'scheduled_operations'):
        try:
            tables.remove(tbl)
        except ValueError:
            LOG.warning('Expected table %(tbl)s was not found in DB.',
                        **locals())
        else:
            tables.append(tbl)

    for table in tables:
        t = Table(table, metadata, autoload=True)
        LOG.info('Purging deleted rows older than age=%(age)d days from '
                 'table=%(table)s', {'age': age_in_days, 'table': table})
        deleted_age = timeutils.utcnow() - dt.timedelta(days=age_in_days)
        try:
            with session.begin():
                result = session.execute(
                    t.delete()
                    .where(t.c.deleted_at < deleted_age))
        except db_exc.DBReferenceError:
            LOG.exception('DBError detected when purging from '
                          'table=%(table)s', {'table': table})
            raise

        rows_purged = result.rowcount
        LOG.info("Deleted %(row)d rows from table=%(table)s",
                 {'row': rows_purged, 'table': table})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from six.moves import http_client

from karbor.api.openstack import ProjectMapper
from karbor.api.v1 import router
from karbor.tests import base
from karbor.tests.unit.api import fakes


class PlansRouterTestCase(base.TestCase):
    def setUp(self):
        super(PlansRouterTestCase, self).setUp()
        mapper = ProjectMapper()
        self.app = router.APIRouter(mapper)

    def test_plans(self):
        req = fakes.HTTPRequest.blank('/fakeproject/plans')
        req.method = 'GET'
        req.content_type = 'application/json'
        response = req.get_response(self.app)
        self.assertEqual(http_client.OK, response.status_int)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Operation classes
"""
import abc
import six

from abc import ABCMeta
from datetime import datetime
from datetime import timedelta
from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import constants
from karbor import context
from karbor import objects
from karbor.services.operationengine import karbor_client


record_operation_log_executor_opts = [
    cfg.IntOpt(
        'retained_operation_log_number',
        default=5,
        help='The number of retained operation log')
]

CONF = cfg.CONF
CONF.register_opts(record_operation_log_executor_opts)

LOG = logging.getLogger(__name__)


@six.add_metaclass(ABCMeta)
class Operation(object):
    OPERATION_TYPE = ""

    def __init__(self, user_trust_manager):
        super(Operation, self).__init__()
        self._user_trust_manager = user_trust_manager
        self._karbor_endpoint = None

    @abc.abstractmethod
    def check_operation_definition(self, operation_definition):
        """Check operation definition

        :param operation_definition: the definition of operation
        """
        pass

    @property
    def karbor_endpoint(self):
        if not self._karbor_endpoint:
            self._karbor_endpoint = karbor_client.get_karbor_endpoint()
        return self._karbor_endpoint

    def run(self, operation_definition, **kwargs):
        param = kwargs.get('param')
        operation_id = param.get('operation_id')
        window = param.get('window_time')
        end_time = param['expect_start_time'] + timedelta(seconds=window)
        is_operation_expired = datetime.utcnow() > end_time

        if constants.OPERATION_RUN_TYPE_RESUME == param['run_type']:
            log_ref = self._get_operation_log(
                operation_id, constants.OPERATION_EXE_STATE_IN_PROGRESS)

            if log_ref is None or len(log_ref) > 1:
                return

            if 1 == len(log_ref):
                log = log_ref[0]
                if is_operation_expired:
                    self._update_log_when_operation_finished(
                        log,
                        constants.OPERATION_EXE_STATE_DROPPED_OUT_OF_WINDOW)
                else:
                    self._resume(operation_definition, param, log)

                self._delete_oldest_operation_log(operation_id)
                return

        if is_operation_expired:
            log_info = {
                'state': constants.OPERATION_EXE_STATE_DROPPED_OUT_OF_WINDOW,
                'end_time': datetime.utcnow()
            }
            log_ref = self._create_operation_log(param, log_info)
        else:
            self._execute(operation_definition, param)

        self._delete_oldest_operation_log(operation_id)

    @abc.abstractmethod
    def _execute(self, operation_definition, param):
        """Execute operation.

        :param operation_definition: the definition of operation
        :param param: dict, other parameters
        """
        pass

    @abc.abstractmethod
    def _resume(self, operation_definition, param, log_ref):
        """Resume operation.

        :param operation_definition: the definition of operation
        :param param: dict, other parameters
        :param log_ref: instance of ScheduledOperationLog
        """
        pass

    def _create_operation_log(self, param, updated_log_info=None):
        log_info = {
            'operation_id': param['operation_id'],
            'expect_start_time': param['expect_start_time'],
            'triggered_time': param['triggered_time'],
            'actual_start_time': datetime.utcnow(),
            'state': constants.OPERATION_EXE_STATE_IN_PROGRESS
        }
        if updated_log_info:
            log_info.update(updated_log_info)

        log_ref = objects.ScheduledOperationLog(context.get_admin_context(),
                                                **log_info)
        try:
            log_ref.create()
        except Exception:
            LOG.exception("Execute operation(%s), create log obj failed",
                          param['operation_id'])
            return
        return log_ref

    def _delete_oldest_operation_log(self, operation_id):
        # delete the oldest logs to keep the number of logs
        # in a reasonable range
        try:
            objects.ScheduledOperationLog.destroy_oldest(
                context.get_admin_context(), operation_id,
                CONF.retained_operation_log_number)
        except Exception:
            pass

    def _update_operation_log(self, log_ref, updates):
        if not log_ref:
            return

        for item in updates:
            setattr(log_ref, item, updates.get(item))
        try:
            log_ref.save()
        except Exception:
            LOG.exception("Execute operation(%s), save log failed",
                          log_ref.operation_id)

    def _update_log_when_operation_finished(self, log_ref, state,
                                            updated_log_info=None):
        if not log_ref:
            return

        updates = {
            'state': state,
            'end_time': datetime.utcnow()
        }
        if updated_log_info:
            updates.update(updated_log_info)

        self._update_operation_log(log_ref, updates)

    def _get_operation_log(self, operation_id, operation_state):
        try:
            logs = objects.ScheduledOperationLogList.get_by_filters(
                context.get_admin_context(),
                {'state': operation_state,
                 'operation_id': operation_id}, limit=2)

            return logs.objects
        except Exception:
            pass

    def _create_karbor_client(self, user_id, project_id):
        token = self._user_trust_manager.get_token(user_id, project_id)
        if not token:
            return None
        ctx = context.get_admin_context()
        ctx.auth_token = token
        ctx.project_id = project_id

        karbor_url = self.karbor_endpoint % {"project_id": project_id}
        return karbor_client.create(ctx, endpoint=karbor_url)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from karbor.common import constants
from karbor import exception
from karbor.i18n import _
from karbor.services.protection import graph
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import uuidutils

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

_INDEX_FILE_NAME = "index.json"
_UUID_STR_LEN = 36


class Checkpoint(object):
    VERSION = "0.9"
    SUPPORTED_VERSIONS = ["0.9"]

    def __init__(self, checkpoint_section, indices_section,
                 bank_lease, checkpoint_id):
        super(Checkpoint, self).__init__()
        self._id = checkpoint_id
        self._checkpoint_section = checkpoint_section
        self._indices_section = indices_section
        self._bank_lease = bank_lease
        self.reload_meta_data()

    def to_dict(self):
        return {
            "id": self.id,
            "status": self.status,
            "protection_plan": self.protection_plan,
            "extra_info": self._md_cache.get("extra_info", None),
            "project_id": self.project_id,
            "resource_graph": self._md_cache.get("resource_graph", None),
            "created_at": self._md_cache.get("created_at", None)
        }

    @property
    def checkpoint_section(self):
        return self._checkpoint_section

    @property
    def id(self):
        return self._id

    @property
    def provider_id(self):
        return self._md_cache["provider_id"]

    @property
    def created_at(self):
        return self._md_cache["created_at"]

    @property
    def status(self):
        # TODO(saggi): check for valid values and transitions
        return self._md_cache["status"]

    @property
    def project_id(self):
        return self._md_cache["project_id"]

    @property
    def owner_id(self):
        # TODO(yinwei): check for valid values and transitions
        return self._md_cache["owner_id"]

    @property
    def resource_graph(self):
        serialized_resource_graph = self._md_cache.get("resource_graph", None)
        if serialized_resource_graph is not None:
            resource_graph = graph.deserialize_resource_graph(
                serialized_resource_graph)
            return resource_graph
        else:
            return None

    @property
    def protection_plan(self):
        return self._md_cache["protection_plan"]

    @status.setter
    def status(self, value):
        self._md_cache["status"] = value

    @resource_graph.setter
    def resource_graph(self, resource_graph):
        serialized_resource_graph = graph.serialize_resource_graph(
            resource_graph)
        self._md_cache["resource_graph"] = serialized_resource_graph

    def _is_supported_version(self, version):
        return version in self.SUPPORTED_VERSIONS

    def _assert_supported_version(self, new_md):
        if new_md["version"] not in self.SUPPORTED_VERSIONS:
            # Something bad happened invalidate the object
            self._md_cache = None
            self._checkpoint_section = None
            raise RuntimeError(
                _("Checkpoint was created in an unsupported version"))

    def reload_meta_data(self):
        try:
            new_md = self._checkpoint_section.get_object(_INDEX_FILE_NAME)
        except exception.BankGetObjectFailed:
            LOG.error("unable to reload metadata for checkpoint id: %s",
                      self.id)
            raise exception.CheckpointNotFound(checkpoint_id=self.id)
        self._assert_supported_version(new_md)
        self._md_cache = new_md

    @classmethod
    def _generate_id(self):
        return uuidutils.generate_uuid()

    @classmethod
    def get_by_section(cls, checkpoints_section, indices_section,
                       bank_lease, checkpoint_id):
        # TODO(yuvalbr) add validation that the checkpoint exists
        checkpoint_section = checkpoints_section.get_sub_section(checkpoint_id)
        return Checkpoint(checkpoint_section, indices_section,
                          bank_lease, checkpoint_id)

    @classmethod
    def create_in_section(cls, checkpoints_section, indices_section,
                          bank_lease, owner_id, plan,
                          checkpoint_id=None, checkpoint_properties=None):
        checkpoint_id = checkpoint_id or cls._generate_id()
        checkpoint_section = checkpoints_section.get_sub_section(checkpoint_id)

        timestamp = timeutils.utcnow_ts()
        created_at = timeutils.utcnow().strftime('%Y-%m-%d')

        provider_id = plan.get("provider_id")
        extra_info = None
        if checkpoint_properties:
            extra_info = checkpoint_properties.get("extra_info", None)
        checkpoint_section.update_object(
            key=_INDEX_FILE_NAME,
            value={
                "version": cls.VERSION,
                "id": checkpoint_id,
                "status": constants.CHECKPOINT_STATUS_PROTECTING,
                "owner_id": owner_id,
                "provider_id": provider_id,
                "project_id": plan.get("project_id"),
                "protection_plan": {
                    "id": plan.get("id"),
                    "name": plan.get("name"),
                    "provider_id": plan.get("provider_id"),
                    "resources": plan.get("resources")
                },
                "extra_info": extra_info,
                "created_at": created_at,
                "timestamp": timestamp
            }
        )

        indices_section.update_object(
            key="/by-provider/%s/%s@%s" % (provider_id, timestamp,
                                           checkpoint_id),
            value=checkpoint_id
        )

        indices_section.update_object(
            key="/by-date/%s/%s@%s" % (created_at, timestamp, checkpoint_id),
            value=checkpoint_id
        )

        indices_section.update_object(
            key="/by-plan/%s/%s/%s@%s" % (
                plan.get("id"), created_at, timestamp, checkpoint_id),
            value=checkpoint_id)

        return Checkpoint(checkpoint_section,
                          indices_section,
                          bank_lease,
                          checkpoint_id)

    def commit(self):
        self._checkpoint_section.update_object(
            key=_INDEX_FILE_NAME,
            value=self._md_cache,
        )

    def purge(self):
        """Purge the index file of the checkpoint.

        Can only be done if the checkpoint has no other files apart from the
        index.
        """
        all_objects = self._checkpoint_section.list_objects()
        if len(all_objects) == 1 and all_objects[0] == _INDEX_FILE_NAME:
            created_at = self._md_cache["created_at"]
            timestamp = self._md_cache["timestamp"]
            plan_id = self._md_cache["protection_plan"]["id"]
            provider_id = self._md_cache["protection_plan"]["provider_id"]
            self._indices_section.delete_object(
                "/by-provider/%s/%s@%s" % (provider_id, timestamp, self.id))
            self._indices_section.delete_object(
                "/by-date/%s/%s@%s" % (created_at, timestamp, self.id))
            self._indices_section.delete_object(
                "/by-plan/%s/%s/%s@%s" % (
                    plan_id, created_at, timestamp, self.id))

            self._checkpoint_section.delete_object(_INDEX_FILE_NAME)
        else:
            raise RuntimeError(_("Could not delete: Checkpoint is not empty"))

    def delete(self):
        self.status = constants.CHECKPOINT_STATUS_DELETED
        self.commit()
        # delete indices
        created_at = self._md_cache["created_at"]
        timestamp = self._md_cache["timestamp"]
        plan_id = self._md_cache["protection_plan"]["id"]
        provider_id = self._md_cache["protection_plan"]["provider_id"]
        self._indices_section.delete_object(
            "/by-provider/%s/%s@%s" % (provider_id, timestamp, self.id))
        self._indices_section.delete_object(
            "/by-date/%s/%s@%s" % (created_at, timestamp, self.id))
        self._indices_section.delete_object(
            "/by-plan/%s/%s/%s@%s" % (
                plan_id, created_at, timestamp, self.id))

    def get_resource_bank_section(self, resource_id):
        prefix = "/resource-data/%s/" % resource_id
        return self._checkpoint_section.get_sub_section(prefix)


class CheckpointCollection(object):

    def __init__(self, bank, bank_lease=None):
        super(CheckpointCollection, self).__init__()
        self._bank = bank
        self._bank_lease = bank_lease
        self._checkpoints_section = bank.get_sub_section("/checkpoints")
        self._indices_section = bank.get_sub_section("/indices")

    def list_ids(self, provider_id, limit=None, marker=None, plan_id=None,
                 start_date=None, end_date=None, sort_dir=None):
        marker_checkpoint = None
        if marker is not None:
            checkpoint_section = self._checkpoints_section.get_sub_section(
                marker)
            marker_checkpoint = checkpoint_section.get_object(_INDEX_FILE_NAME)
            timestamp = marker_checkpoint["timestamp"]
            marker = "%s@%s" % (timestamp, marker)

        if start_date is not None:
            if end_date is None:
                end_date = timeutils.utcnow()

        if plan_id is None and start_date is None:
            prefix = "/by-provider/%s/" % provider_id
            if marker is not None:
                marker = "/%s" % marker
        elif plan_id is not None:
            prefix = "/by-plan/%s/" % plan_id
            if marker is not None:
                date = marker_checkpoint["created_at"]
                marker = "/by-plan/%s/%s/%s" % (plan_id, date, marker)
        else:
            prefix = "/by-date/"
            if marker is not None:
                date = marker_checkpoint["created_at"]
                marker = "/by-date/%s/%s" % (date, marker)

        return self._list_ids(prefix, limit, marker, start_date, end_date,
                              sort_dir)

    def _list_ids(self, prefix, limit, marker, start_date, end_date, sort_dir):
        if start_date is None:
            return [key[key.find("@") + 1:]
                    for key in self._indices_section.list_objects(
                    prefix=prefix,
                    limit=limit,
                    marker=marker,
                    sort_dir=sort_dir
                    )]
        else:
            ids = []
            for key in self._indices_section.list_objects(prefix=prefix,
                                                          marker=marker,
                                                          sort_dir=sort_dir):
                date = datetime.strptime(key.split("/")[-2], "%Y-%m-%d")
                if start_date <= date <= end_date:
                    ids.append(key[key.find("@") + 1:])
                if limit is not None and len(ids) == limit:
                    return ids
            return ids

    def get(self, checkpoint_id):
        # TODO(saggi): handle multiple instances of the same checkpoint
        return Checkpoint.get_by_section(self._checkpoints_section,
                                         self._indices_section,
                                         self._bank_lease,
                                         checkpoint_id)

    def create(self, plan, checkpoint_properties=None):
        # TODO(saggi): Serialize plan to checkpoint. Will be done in
        # future patches.
        return Checkpoint.create_in_section(
            self._checkpoints_section,
            self._indices_section,
            self._bank_lease,
            self._bank.get_owner_id(),
            plan,
            checkpoint_properties=checkpoint_properties)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from karbor import db
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor.objects import base


@base.KarborObjectRegistry.register
class ScheduledOperationState(base.KarborPersistentObject, base.KarborObject,
                              base.KarborObjectDictCompat,
                              base.KarborComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.IntegerField(),
        'operation_id': fields.UUIDField(),
        'service_id': fields.IntegerField(),
        'trust_id': fields.StringField(),
        'state': fields.StringField(),
        'end_time_for_run': base.DateTimeField(nullable=True),

        'operation': fields.ObjectField("ScheduledOperation")
    }

    INSTANCE_OPTIONAL_JOINED_FIELDS = ['operation']

    @staticmethod
    def _from_db_object(context, state, db_state, expected_attrs=[]):
        special_fields = set(state.INSTANCE_OPTIONAL_JOINED_FIELDS)
        normal_fields = set(state.fields) - special_fields
        for name in normal_fields:
            state[name] = db_state.get(name)

        if 'operation' in expected_attrs:
            if db_state.get('operation', None) is None:
                state.operation = None
            else:
                if not state.obj_attr_is_set('operation'):
                    state.operation = objects.ScheduledOperation(context)
                state.operation._from_db_object(context, state.operation,
                                                db_state['operation'])

        state._context = context
        state.obj_reset_changes()
        return state

    @base.remotable_classmethod
    def get_by_operation_id(cls, context, operation_id, expected_attrs=[]):
        columns_to_join = [col for col in expected_attrs
                           if col in cls.INSTANCE_OPTIONAL_JOINED_FIELDS]

        db_state = db.scheduled_operation_state_get(
            context, operation_id, columns_to_join)
        if db_state:
            return cls._from_db_object(context, cls(),
                                       db_state, columns_to_join)

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))

        updates = self.karbor_obj_get_changes()
        db_state = db.scheduled_operation_state_create(self._context, updates)
        self._from_db_object(self._context, self, db_state)

    @base.remotable
    def save(self):
        updates = self.karbor_obj_get_changes()
        if updates and self.operation_id:
            db.scheduled_operation_state_update(self._context,
                                                self.operation_id,
                                                updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        if self.operation_id:
            db.scheduled_operation_state_delete(self._context,
                                                self.operation_id)


@base.KarborObjectRegistry.register
class ScheduledOperationStateList(base.ObjectListBase, base.KarborObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('ScheduledOperationState'),
    }

    @base.remotable_classmethod
    def get_by_filters(cls, context, filters, limit=None, marker=None,
                       sort_keys=None, sort_dirs=None, columns_to_join=[]):

        option_column = ScheduledOperationState.INSTANCE_OPTIONAL_JOINED_FIELDS
        valid_columns = [column for column in columns_to_join
                         if column in option_column]

        db_state_list = db.scheduled_operation_state_get_all_by_filters_sort(
            context, filters, limit=limit, marker=marker, sort_keys=sort_keys,
            sort_dirs=sort_dirs, columns_to_join=valid_columns)

        return base.obj_make_list(
            context, cls(context), ScheduledOperationState, db_state_list,
            expected_attrs=valid_columns)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock
from oslo_config import cfg

from karbor.api.v1 import protectables
from karbor import context
from karbor import exception
from karbor.tests import base
from karbor.tests.unit.api import fakes

CONF = cfg.CONF


class ProtectablesApiTest(base.TestCase):
    def setUp(self):
        super(ProtectablesApiTest, self).setUp()
        self.controller = protectables.ProtectablesController()
        self.ctxt = context.RequestContext('demo', 'fakeproject', True)

    @mock.patch(
        'karbor.api.v1.protectables.ProtectablesController._get_all')
    def test_protectables_list_detail(self, moak_get_all):
        req = fakes.HTTPRequest.blank('/v1/protectables')
        self.controller.index(req)
        self.assertTrue(moak_get_all.called)

    @mock.patch(
        'karbor.services.protection.api.API.show_protectable_type')
    @mock.patch(
        'karbor.api.v1.protectables.ProtectablesController._get_all')
    def test_protectables_show(
            self, moak_get_all, moak_show_protectable_type):
        req = fakes.HTTPRequest.blank('/v1/protectables')
        moak_get_all.return_value = ["OS::Keystone::Project"]
        self.controller.show(req, 'OS::Keystone::Project')
        self.assertTrue(moak_get_all.called)
        self.assertTrue(moak_show_protectable_type.called)

    @mock.patch(
        'karbor.api.v1.protectables.ProtectablesController._get_all')
    def test_protectables_show_Invalid(self, moak_get_all):
        req = fakes.HTTPRequest.blank('/v1/protectables')
        moak_get_all.return_value = ["OS::Keystone::Project"]
        self.assertRaises(exception.InvalidInput, self.controller.show,
                          req, "1")
        self.assertTrue(moak_get_all.called)

    @mock.patch(
        'karbor.services.protection.api.API.'
        'list_protectable_instances')
    @mock.patch(
        'karbor.api.v1.protectables.ProtectablesController._get_all')
    def test_protectables_instances_index(self, moak_get_all,
                                          moak_list_protectable_instances):
        req = fakes.HTTPRequest.blank('/v1/protectables')
        moak_get_all.return_value = ["OS::Keystone::Project"]
        self.controller.instances_index(req, 'OS::Keystone::Project')
        self.assertTrue(moak_get_all.called)
        self.assertTrue(moak_list_protectable_instances.called)

    @mock.patch(
        'karbor.services.protection.api.API.'
        'list_protectable_dependents')
    @mock.patch(
        'karbor.services.protection.api.API.'
        'show_protectable_instance')
    @mock.patch(
        'karbor.api.v1.protectables.ProtectablesController._get_all')
    def test_protectables_instances_show(self, moak_get_all,
                                         moak_show_protectable_instance,
                                         moak_list_protectable_dependents):
        req = fakes.HTTPRequest.blank('/v1/protectables')
        moak_get_all.return_value = ["OS::Keystone::Project"]
        self.controller.instances_show(
            req,
            'OS::Keystone::Project',
            'efc6a88b-9096-4bb6-8634-cda182a6e12a',
        )
        self.assertTrue(moak_get_all.called)
        self.assertTrue(moak_show_protectable_instance.called)
        self.assertTrue(moak_list_protectable_dependents.called)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from abclient import client
from oslo_config import cfg
from oslo_log import log as logging

from karbor import utils

EISOO_JOB_TYPE = (ORACLE_JOB_TYPE) = (1)
EISOO_JOB_STATUS = (RUNNING, SUCCESS, FAILED) = (4, 32, 64)

SERVICE = "eisoo"
eisoo_client_opts = [
    cfg.StrOpt(SERVICE + '_endpoint',
               help='URL of the eisoo endpoint.'),
    cfg.StrOpt(SERVICE + '_app_id',
               help='App id for eisoo authentication.'),
    cfg.StrOpt(SERVICE + '_app_secret',
               secret=True,
               help='App secret for eisoo authentication.'),
]

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def create(context, conf):
    config_dir = utils.find_config(CONF.provider_config_dir)
    config_file = os.path.abspath(os.path.join(config_dir,
                                               'eisoo.conf'))
    config = cfg.ConfigOpts()
    config(args=['--config-file=' + config_file])
    config.register_opts(eisoo_client_opts,
                         group=SERVICE + '_client')

    LOG.info('Creating eisoo client with url %s.',
             config.eisoo_client.eisoo_endpoint)
    abclient = client.ABClient(config.eisoo_client.eisoo_endpoint,
                               config.eisoo_client.eisoo_app_id,
                               config.eisoo_client.eisoo_app_secret)

    return abclient



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock
from oslo_config import cfg
from webob import exc

from karbor.api.v1 import plans
from karbor.common import constants
from karbor import context
from karbor import exception
from karbor.tests import base
from karbor.tests.unit.api import fakes

CONF = cfg.CONF

DEFAULT_NAME = 'My 3 tier application'
DEFAULT_DESCRIPTION = 'My 3 tier application protection plan'
DEFAULT_PROVIDER_ID = 'efc6a88b-9096-4bb6-8634-cda182a6e12a'
DEFAULT_PROJECT_ID = '39bb894794b741e982bd26144d2949f6'
DEFAULT_RESOURCES = [{'id': 'efc6a88b-9096-4bb6-8634-cda182a6e144',
                      "type": "OS::Cinder::Volume", "name": "name1"}]
DEFAULT_PARAMETERS = {"OS::Cinder::Volume": {"backup_name": "name"}}


class PlanApiTest(base.TestCase):
    def setUp(self):
        super(PlanApiTest, self).setUp()
        self.controller = plans.PlansController()
        self.ctxt = context.RequestContext('demo', 'fakeproject', True)

    @mock.patch(
        'karbor.services.protection.rpcapi.ProtectionAPI.show_provider')
    @mock.patch(
        'karbor.objects.plan.Plan.create')
    def test_plan_create(self, mock_plan_create, mock_provider):
        plan = self._plan_in_request_body()
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        mock_provider.return_value = fakes.PROVIDER_OS
        self.controller.create(req, body)
        self.assertTrue(mock_plan_create.called)

    def test_plan_create_InvalidBody(self):
        plan = self._plan_in_request_body()
        body = {"planxx": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.assertRaises(exc.HTTPUnprocessableEntity, self.controller.create,
                          req, body)

    def test_plan_create_InvalidProviderId(self):
        plan = self._plan_in_request_body(
            name=DEFAULT_NAME,
            description=DEFAULT_DESCRIPTION,
            provider_id="",
            status=constants.PLAN_STATUS_SUSPENDED,
            project_id=DEFAULT_PROJECT_ID,
            resources=[])
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.assertRaises(exception.InvalidInput, self.controller.create,
                          req, body)

    def test_plan_create_InvalidResources(self):
        plan = self._plan_in_request_body(
            name=DEFAULT_NAME,
            description=DEFAULT_DESCRIPTION,
            provider_id=DEFAULT_PROVIDER_ID,
            status=constants.PLAN_STATUS_SUSPENDED,
            project_id=DEFAULT_PROJECT_ID,
            resources=[])
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.assertRaises(exception.InvalidInput, self.controller.create,
                          req, body)

    @mock.patch(
        'karbor.services.protection.rpcapi.ProtectionAPI.show_provider')
    def test_plan_create_InvalidParameters(self, mock_provider):
        parameters = {"OS::Cinder::Volume": {"test": "os"}}
        plan = self._plan_in_request_body(
            name=DEFAULT_NAME,
            description=DEFAULT_DESCRIPTION,
            provider_id=DEFAULT_PROVIDER_ID,
            status=constants.PLAN_STATUS_SUSPENDED,
            project_id=DEFAULT_PROJECT_ID,
            parameters=parameters)
        body = {"plan": plan}
        mock_provider.return_value = fakes.PROVIDER_OS
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.assertRaises(exc.HTTPBadRequest, self.controller.create,
                          req, body)

    @mock.patch(
        'karbor.api.v1.plans.PlansController._plan_get')
    @mock.patch(
        'karbor.api.v1.plans.PlansController._plan_update')
    def test_plan_update(self, mock_plan_update, mock_plan_get):
        plan = self._plan_in_request_body()
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.controller.update(
            req, "2a9ce1f3-cc1a-4516-9435-0ebb13caa398", body)
        self.assertTrue(mock_plan_update.called)
        self.assertTrue(mock_plan_get.called)

    def test_plan_update_InvalidBody(self):
        plan = self._plan_in_request_body()
        body = {"planxx": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.assertRaises(
            exc.HTTPBadRequest, self.controller.update,
            req, "2a9ce1f3-cc1a-4516-9435-0ebb13caa398", body)

    def test_plan_update_InvalidId(self):
        plan = self._plan_in_request_body()
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.assertRaises(
            exc.HTTPNotFound, self.controller.update,
            req, "2a9ce1f3-cc1a-4516-9435-0ebb13caa398", body)

    def test_plan_update_InvalidResources(self):
        plan = self._plan_in_request_body(
            name=DEFAULT_NAME,
            description=DEFAULT_DESCRIPTION,
            provider_id=DEFAULT_PROVIDER_ID,
            status=constants.PLAN_STATUS_SUSPENDED,
            project_id=DEFAULT_PROJECT_ID,
            resources=[{'key1': 'value1'}])
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.assertRaises(
            exception.InvalidInput, self.controller.update,
            req, "2a9ce1f3-cc1a-4516-9435-0ebb13caa398", body)

    @mock.patch(
        'karbor.api.v1.plans.PlansController._get_all')
    def test_plan_list_detail(self, moak_get_all):
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.controller.index(req)
        self.assertTrue(moak_get_all.called)

    @mock.patch(
        'karbor.api.v1.plans.PlansController._get_all')
    def test_plan_index_limit_offset(self, moak_get_all):
        req = fakes.HTTPRequest.blank('/v1/plans?limit=2&offset=1')
        self.controller.index(req)
        self.assertTrue(moak_get_all.called)

        req = fakes.HTTPRequest.blank('/v1/plans?limit=-1&offset=1')
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.index,
                          req)

        req = fakes.HTTPRequest.blank('/v1/plans?limit=a&offset=1')
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.index,
                          req)

        url = '/v1/plans?limit=2&offset=43543564546567575'
        req = fakes.HTTPRequest.blank(url)
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.index,
                          req)

    def test_plan_create_empty_dict(self):
        plan = self._plan_in_request_body(parameters={})
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.controller.create(req, body)

    @mock.patch(
        'karbor.api.v1.plans.PlansController._plan_get')
    def test_plan_show(self, moak_plan_get):
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.controller.show(req, '2a9ce1f3-cc1a-4516-9435-0ebb13caa398')
        self.assertTrue(moak_plan_get.called)

    def test_plan_show_Invalid(self):
        req = fakes.HTTPRequest.blank('/v1/plans/1')
        self.assertRaises(
            exc.HTTPBadRequest, self.controller.show,
            req, "1")

    @mock.patch(
        'karbor.api.v1.plans.PlansController._plan_get')
    def test_plan_delete(self, moak_plan_get):
        req = fakes.HTTPRequest.blank('/v1/plans')
        self.controller.delete(req, '2a9ce1f3-cc1a-4516-9435-0ebb13caa398')
        self.assertTrue(moak_plan_get.called)

    def test_plan_delete_Invalid(self):
        req = fakes.HTTPRequest.blank('/v1/plans/1')
        self.assertRaises(
            exc.HTTPBadRequest, self.controller.delete,
            req, "1")

    @mock.patch(
        'karbor.api.v1.plans.check_policy')
    @mock.patch(
        'karbor.api.v1.plans.PlansController._plan_get')
    def test_plan_update_InvalidStatus(
            self, mock_plan_get, mock_check_policy):
        plan = self._plan_in_request_body(
            name=DEFAULT_NAME,
            description=DEFAULT_DESCRIPTION,
            provider_id=DEFAULT_PROVIDER_ID,
            status=constants.PLAN_STATUS_STARTED,
            project_id=DEFAULT_PROJECT_ID,
            resources=DEFAULT_RESOURCES)
        body = {"plan": plan}
        req = fakes.HTTPRequest.blank('/v1/plans')
        mock_plan_get.return_value = plan
        self.assertRaises(exception.InvalidPlan,
                          self.controller.update, req,
                          "2a9ce1f3-cc1a-4516-9435-0ebb13caa398",
                          body)

    def _plan_in_request_body(self, name=DEFAULT_NAME,
                              description=DEFAULT_DESCRIPTION,
                              provider_id=DEFAULT_PROVIDER_ID,
                              status=constants.PLAN_STATUS_SUSPENDED,
                              project_id=DEFAULT_PROJECT_ID,
                              resources=DEFAULT_RESOURCES,
                              parameters=DEFAULT_PARAMETERS):
        plan_req = {
            'name': name,
            'description': description,
            'provider_id': provider_id,
            'status': status,
            'project_id': project_id,
            'resources': resources,
            'parameters': parameters,
        }

        return plan_req



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from karbor.common import constants
from karbor import exception
from karbor import resource
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protectable_plugin
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

INVALID_IMAGE_STATUS = ['killed', 'deleted', 'pending_delete',
                        'deactivated']


class ImageProtectablePlugin(protectable_plugin.ProtectablePlugin):
    """Glance image protectable plugin"""
    _SUPPORT_RESOURCE_TYPE = constants.IMAGE_RESOURCE_TYPE

    def _glance_client(self, context):
        self._glance_client_instance = ClientFactory.create_client(
            'glance', context)
        return self._glance_client_instance

    def _nova_client(self, context):
        self._nova_client_instance = ClientFactory.create_client(
            'nova', context)
        return self._nova_client_instance

    def get_resource_type(self):
        return self._SUPPORT_RESOURCE_TYPE

    def get_parent_resource_types(self):
        return (constants.SERVER_RESOURCE_TYPE,
                constants.PROJECT_RESOURCE_TYPE,)

    def list_resources(self, context, parameters=None):
        try:
            images = self._glance_client(context).images.list()
        except Exception as e:
            LOG.exception("List all images from glance failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                      id=image.id, name=image.name)
                    for image in images
                    if image.status not in INVALID_IMAGE_STATUS]

    def _get_dependent_resources_by_server(self,
                                           context,
                                           parent_resource):
        try:
            server = self._nova_client(context).servers.get(parent_resource.id)
        except Exception as e:
            LOG.exception("List all server from nova failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))

        if not server.image:
            return []
        try:
            image = self._glance_client(context).images.get(server.image['id'])
        except Exception as e:
            LOG.exception("Getting image from glance failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                  id=server.image['id'],
                                  name=image.name)]

    def _get_dependent_resources_by_project(self,
                                            context,
                                            parent_resource):
        try:
            images = self._glance_client(context).images.list()
        except Exception as e:
            LOG.exception("List all images from glance failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                      id=image.id,
                                      name=image.name)
                    for image in images
                    if image.owner == parent_resource.id
                    and image.status not in INVALID_IMAGE_STATUS]

    def show_resource(self, context, resource_id, parameters=None):
        try:
            image = self._glance_client(context).images.get(resource_id)
        except Exception as e:
            LOG.exception("Show a image from glance failed.")
            raise exception.ProtectableResourceNotFound(
                id=resource_id,
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            if image.status in INVALID_IMAGE_STATUS:
                raise exception.ProtectableResourceInvalidStatus(
                    id=image.id, type=self._SUPPORT_RESOURCE_TYPE,
                    status=image.status)
            return resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                     id=image.id, name=image.name)

    def get_dependent_resources(self, context, parent_resource):
        if parent_resource.type == constants.SERVER_RESOURCE_TYPE:
            return self._get_dependent_resources_by_server(context,
                                                           parent_resource)

        if parent_resource.type == constants.PROJECT_RESOURCE_TYPE:
            return self._get_dependent_resources_by_project(context,
                                                            parent_resource)

        return []



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""karbor base exception handling.

Includes decorator for re-raising karbor-type exceptions.

SHOULD include dedicated exception logging.

"""

import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_versionedobjects import exception as obj_exc
import six
import webob.exc
from webob.util import status_generic_reasons
from webob.util import status_reasons

from six.moves import http_client

from karbor.i18n import _


LOG = logging.getLogger(__name__)

exc_log_opts = [
    cfg.BoolOpt('fatal_exception_format_errors',
                default=False,
                help='Make exception message format errors fatal.'),
]

CONF = cfg.CONF
CONF.register_opts(exc_log_opts)


class ConvertedException(webob.exc.WSGIHTTPException):
    def __init__(self, code=500, title="",
                 explanation=""):
        self.code = code
        # There is a strict rule about constructing status line for HTTP:
        # '...Status-Line, consisting of the protocol version followed by a
        # numeric status code and its associated textual phrase, with each
        # element separated by SP characters'
        # (http://www.faqs.org/rfcs/rfc2616.html)
        # 'code' and 'title' can not be empty because they correspond
        # to numeric status code and its associated text
        if title:
            self.title = title
        else:
            try:
                self.title = status_reasons[self.code]
            except KeyError:
                generic_code = self.code // 100
                self.title = status_generic_reasons[generic_code]
        self.explanation = explanation
        super(ConvertedException, self).__init__()


class Error(Exception):
    pass


class KarborException(Exception):
    """Base karbor Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.

    """
    message = _("An unknown exception occurred.")
    code = http_client.INTERNAL_SERVER_ERROR
    headers = {}
    safe = True

    def __init__(self, message=None, **kwargs):
        """Initiate the instance of KarborException

        There are two ways to initiate the instance.
        1. Specify the value of 'message' and leave the 'kwargs' None.
        2. Leave 'message' None, and specify the keyword arguments matched
           with the format of KarborException.message. Especially, can't
           use the 'message' as the key in the 'kwargs', otherwise, the
           first argument('message') will be set.

        Note: This class doesn't support to create instance of KarborException
            with another instance.
        """
        self.kwargs = kwargs

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            try:
                message = self.message % kwargs

            except Exception:
                exc_info = sys.exc_info()
                # kwargs doesn't match a variable in the message
                # log the issue and the kwargs
                LOG.exception('Exception in string format operation')
                for name, value in kwargs.items():
                    LOG.error("%(name)s: %(value)s",
                              {'name': name, 'value': value})
                if CONF.fatal_exception_format_errors:
                    six.reraise(*exc_info)
                # at least get the core message out if something happened
                message = self.message
        elif isinstance(message, Exception):
            message = six.text_type(message)

        # NOTE(luisg): We put the actual message in 'msg' so that we can access
        # it, because if we try to access the message via 'message' it will be
        # overshadowed by the class' message attribute
        self.msg = message
        super(KarborException, self).__init__(message)

    def __unicode__(self):
        return self.msg


class NotAuthorized(KarborException):
    message = _("Not authorized.")
    code = http_client.FORBIDDEN


class AdminRequired(NotAuthorized):
    message = _("User does not have admin privileges")


class PolicyNotAuthorized(NotAuthorized):
    message = _("Policy doesn't allow %(action)s to be performed.")


class AuthorizationFailure(NotAuthorized):
    message = _("Authorization for %(obj)s is failed ")


class Invalid(KarborException):
    message = _("Unacceptable parameters.")
    code = http_client.BAD_REQUEST


class InvalidParameterValue(Invalid):
    message = _("%(err)s")


class InvalidInput(Invalid):
    message = _("Invalid input received: %(reason)s")


class ScheduledOperationExist(Invalid):
    message = _("Scheduled Operation%(op_id)s exists")


class NotFound(KarborException):
    message = _("Resource could not be found.")
    code = http_client.NOT_FOUND
    safe = True


class ConfigNotFound(NotFound):
    message = _("Could not find config at %(path)s")


class MalformedRequestBody(KarborException):
    message = _("Malformed message body: %(reason)s")


class InvalidContentType(Invalid):
    message = _("Invalid content type %(content_type)s.")


class InvalidProtectableInstance(Invalid):
    message = _("Invalid protectable instance.")


class PasteAppNotFound(NotFound):
    message = _("Could not load paste app '%(name)s' from %(path)s")


class ServiceNotFound(NotFound):
    message = _("Service %(service_id)s could not be found.")


class HostBinaryNotFound(NotFound):
    message = _("Could not find binary %(binary)s on host %(host)s.")


class TriggerNotFound(NotFound):
    message = _("Trigger %(id)s could not be found.")


class ScheduledOperationNotFound(NotFound):
    message = _("Scheduled Operation %(id)s could not be found.")


class ScheduledOperationStateNotFound(NotFound):
    message = _("Scheduled Operation State %(op_id)s could not be found.")


class ScheduledOperationLogNotFound(NotFound):
    message = _("Scheduled Operation Log %(log_id)s could not be found.")


class ListProtectableResourceFailed(KarborException):
    message = _("List protectable resources of type %(type)s failed: "
                "%(reason)s")


class ProtectableResourceNotFound(NotFound):
    message = _("The resource %(id)s of type %(type)s could not be found: "
                "%(reason)s")


class ProtectableResourceInvalidStatus(KarborException):
    message = _("The resource %(id)s of type %(type)s has a invalid "
                "status: %(status)s")


class InvalidOperationObject(Invalid):
    message = _("The operation %(operation_id)s is invalid")


class DeleteTriggerNotAllowed(NotAuthorized):
    message = _("Can not delete trigger %(trigger_id)s")


class ClassNotFound(NotFound):
    message = _("Class %(class_name)s could not be found: %(exception)s")


class InvalidOperationDefinition(Invalid):
    message = _("Invalid operation definition, reason:%(reason)s")


OrphanedObjectError = obj_exc.OrphanedObjectError
ObjectActionError = obj_exc.ObjectActionError


class PlanNotFound(NotFound):
    message = _("Plan %(plan_id)s could not be found.")


class RestoreNotFound(NotFound):
    message = _("Restore %(restore_id)s could not be found.")


class OperationLogNotFound(NotFound):
    message = _("OperationLog %(restore_id)s could not be found.")


class InvalidPlan(Invalid):
    message = _("Invalid plan: %(reason)s")


class ProtectableTypeNotFound(NotFound):
    message = _("ProtectableType %(protectable_type)s could"
                " not be found.")


class ProtectionPluginNotFound(NotFound):
    message = _("Protection Plugin for %(type)s could"
                " not be found.")


class ProviderNotFound(NotFound):
    message = _("Provider %(provider_id)s could"
                " not be found.")


class CheckpointRecordNotFound(NotFound):
    message = _("CheckpointRecord %(id)s could not be found.")


class CreateBackupFailed(KarborException):
    message = _("Create Backup failed: %(reason)s, id=%(resource_id)s,"
                " type=%(resource_type)s")


class CreateResourceFailed(KarborException):
    message = _("Create %(name)s failed: %(reason)s, id=%(resource_id)s,"
                " type=%(resource_type)s")


class DeleteResourceFailed(KarborException):
    message = _("Delete %(name)s failed: %(reason)s, id=%(resource_id)s,"
                " type=%(resource_type)s")


class RestoreResourceFailed(KarborException):
    message = _("Restore %(name)s failed: %(reason)s, id=%(resource_id)s,"
                " type=%(resource_type)s")


class DeleteBackupFailed(KarborException):
    message = _("Delete Backup failed: %(reason)s, id=%(resource_id)s,"
                " type=%(resource_type)s")


class RestoreBackupFailed(KarborException):
    message = _("Restore Backup failed: %(reason)s, id=%(resource_id)s,"
                " type=%(resource_type)s")


class GetBackupFailed(KarborException):
    message = _("Get Backup failed: %(reason)s, id=%(resource_id)s,"
                " type=%(resource_type)s")


class FlowError(KarborException):
    message = _("Flow: %(flow)s, Error: %(error)s")


class CheckpointNotFound(NotFound):
    message = _("Checkpoint %(checkpoint_id)s could"
                " not be found.")


class BankCreateObjectFailed(KarborException):
    message = _("Create Object in Bank Failed: %(reason)s")


class BankUpdateObjectFailed(KarborException):
    message = _("Update Object %(key)s in Bank Failed: %(reason)s")


class BankDeleteObjectFailed(KarborException):
    message = _("Delete Object %(key)s in Bank Failed: %(reason)s")


class BankGetObjectFailed(KarborException):
    message = _("Get Object %(key)s in Bank Failed: %(reason)s")


class BankListObjectsFailed(KarborException):
    message = _("Get Object in Bank Failed: %(reason)s")


class BankReadonlyViolation(KarborException):
    message = _("Bank read-only violation")


class AcquireLeaseFailed(KarborException):
    message = _("Acquire Lease in Failed: %(reason)s")


class CreateContainerFailed(KarborException):
    message = _("Create Container in Bank Failed: %(reason)s")


class CreateBucketFailed(KarborException):
    message = _("Create Bucket in Bank Failed: %(reason)s")


class TriggerIsInvalid(Invalid):
    message = _("Trigger%(trigger_id)s is invalid.")


class InvalidTaskFlowObject(Invalid):
    message = _("The task flow is invalid: %(reason)s")


class InvalidOriginalId(Invalid):
    message = _("The original_id: %(original_id)s is invalid.")


class CheckpointNotAvailable(KarborException):
    message = _("The checkpoint %(checkpoint_id)s is not available")


class CheckpointNotBeDeleted(KarborException):
    message = _("The checkpoint %(checkpoint_id)s can not be deleted.")


class GetProtectionNetworkSubResourceFailed(KarborException):
    message = _("Get protection network sub-resources of type %(type)s failed:"
                " %(reason)s")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import timeutils

from karbor import context
from karbor import objects
from karbor.tests.unit import objects as test_objects

NOW = timeutils.utcnow().replace(microsecond=0)

Operation_ID = '0354ca9ddcd046b693340d78759fd274'

Fake_State = {
    'created_at': NOW,
    'deleted_at': None,
    'updated_at': NOW,
    'deleted': False,
    'id': 1,
    'operation_id': Operation_ID,
    'service_id': 2,
    'trust_id': '123',
    'state': 'triggered',
}


class TestScheduledOperationState(test_objects.BaseObjectsTestCase):
    State_Class = objects.ScheduledOperationState

    @mock.patch('karbor.db.scheduled_operation_state_get')
    def test_get_by_operation_id(self, state_get):
        db_state = Fake_State
        state_get.return_value = db_state
        state = self.State_Class.get_by_operation_id(self.context,
                                                     Operation_ID)
        self._compare(self, db_state, state)
        state_get.assert_called_once_with(self.context, Operation_ID, [])

    @mock.patch('karbor.db.scheduled_operation_state_create')
    def test_create(self, state_create):
        db_state = Fake_State
        state_create.return_value = db_state
        state = self.State_Class(context=self.context)
        state.create()
        self._compare(self, db_state, state)
        state_create.assert_called_once_with(self.context, {})

    @mock.patch('karbor.db.scheduled_operation_state_update')
    def test_save(self, state_update):
        db_state = Fake_State
        state = self.State_Class._from_db_object(self.context,
                                                 self.State_Class(),
                                                 db_state)
        state.state = 'triggered'
        state.save()

        state_update.assert_called_once_with(self.context,
                                             state.operation_id,
                                             {'state': 'triggered'})

    @mock.patch('karbor.db.scheduled_operation_state_delete')
    def test_destroy(self, state_delete):
        db_state = Fake_State
        state = self.State_Class._from_db_object(self.context,
                                                 self.State_Class(),
                                                 db_state)
        state.destroy()
        state_delete.assert_called_once_with(self.context,
                                             state.operation_id)

    def test_get_state_and_operation(self):
        ctx = context.get_admin_context()
        service, trigger, operation, state = FakeEnv(ctx).do_init()

        state_obj = self.State_Class.get_by_operation_id(
            self.context, operation.id, ['operation'])

        self.assertEqual(operation.id, state_obj.operation.id)


class TestScheduledOperationStateList(test_objects.BaseObjectsTestCase):

    def setUp(self):
        super(TestScheduledOperationStateList, self).setUp()
        self.context = context.get_admin_context()

    def test_get_by_filters(self):
        service, trigger, operation, state = FakeEnv(self.context).do_init()
        states = objects.ScheduledOperationStateList.get_by_filters(
            self.context, {'service_id': service.id},
            columns_to_join=['operation'])
        self.assertEqual(1, len(states.objects))
        state1 = states.objects[0]
        self.assertEqual(state.id, state1.id)
        self.assertEqual(operation.id, state1.operation.id)


class FakeEnv(object):

    def __init__(self, ctx):
        super(FakeEnv, self).__init__()
        self.context = ctx

    def do_init(self):
        service = self._create_service()
        trigger = self._create_trigger()
        operation = self._create_operation(trigger.id)
        state = self._create_operation_state(operation.id, service.id)
        return service, trigger, operation, state

    def _create_service(self):
        service_info = {
            'host': "abc",
            'binary': 'karbor-operationengine'
        }
        service = objects.Service(self.context, **service_info)
        service.create()
        return service

    def _create_trigger(self):
        trigger_info = {
            'name': 'daily',
            'project_id': '123',
            'type': 'time',
            'properties': {}
        }
        trigger = objects.Trigger(self.context, **trigger_info)
        trigger.create()
        return trigger

    def _create_operation(self, trigger_id):
        operation_info = {
            'name': 'protect vm',
            'description': 'protect vm resource',
            'operation_type': 'protect',
            'user_id': '123',
            'project_id': '123',
            'trigger_id': trigger_id,
            'operation_definition': {}
        }
        operation = objects.ScheduledOperation(self.context, **operation_info)
        operation.create()
        return operation

    def _create_operation_state(self, operation_id, service_id):
        state_info = {
            'operation_id': operation_id,
            'service_id': service_id,
            'trust_id': '123',
            'state': 'triggered',
        }
        state = objects.ScheduledOperationState(self.context, **state_info)
        state.create()
        return state



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor import exception
from karbor.i18n import _


def _parse_service_catalog_info(config, context):
    try:
        service_type, service_name, endpoint_type = config.split(':')
    except ValueError:
        msg = _("Failed to parse the catalog info option %s, "
                "must be in the form: "
                "<service_type>:<service_name>:<endpoint_type>"
                ) % config
        raise exception.KarborException(msg)

    for entry in context.service_catalog:
        if entry.get('type') == service_type:
            return entry.get('endpoints')[0].get(endpoint_type)


def _parse_service_endpoint(endpoint_url, context, append_project_fmt=None):
    if not endpoint_url:
        return None

    if not append_project_fmt:
        return endpoint_url

    return append_project_fmt % {
        'url': endpoint_url,
        'project': context.project_id,
    }


def get_url(service, context, client_config,
            append_project_fmt=None, **kwargs):
    '''Return the url of given service endpoint.'''

    url = ""
    privileged_user = kwargs.get('privileged_user')
    # get url by endpoint
    if privileged_user is not True:
        try:
            url = _parse_service_endpoint(
                getattr(client_config, '%s_endpoint' % service),
                context, append_project_fmt)
            if url:
                return url
        except Exception:
            pass

        # get url by catalog
        try:
            url = _parse_service_catalog_info(
                getattr(client_config, '%s_catalog_info' % service), context)
            if url:
                return url
        except Exception:
            pass

    # get url by accessing keystone
    try:
        keystone_plugin = kwargs.get('keystone_plugin')
        url = keystone_plugin.get_service_endpoint(
            client_config.service_name, client_config.service_type,
            client_config.region_id, client_config.interface)

        url = url.replace("$", "%")
    except Exception:
        pass

    if url:
        return url

    raise exception.KarborException(
        _("Couldn't find the endpoint of service(%s)") % service)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor.services.protection.flows import workflow
from karbor.tests import base


def fake_func():
    return True


class WorkFlowTest(base.TestCase):
    def setUp(self):
        super(WorkFlowTest, self).setUp()
        self.workflow_engine = workflow.TaskFlowEngine()

    def test_create_task(self):
        test_task = self.workflow_engine.create_task(fake_func, name='fake')
        self.assertEqual('fake', test_task.name)

    def test_add_task(self):
        test_flow = self.workflow_engine.build_flow('test')
        test_task = self.workflow_engine.create_task(fake_func, name='fake')
        self.workflow_engine.add_tasks(test_flow, test_task)
        self.assertEqual(1, len(test_flow))

    def test_search_task(self):
        flow = self.workflow_engine.build_flow('test')
        task1 = self.workflow_engine.create_task(fake_func, name='fake_func')
        task2 = self.workflow_engine.create_task(fake_func, name='fake_func2')
        self.workflow_engine.add_tasks(flow, task1, task2)
        result = self.workflow_engine.search_task(flow, 'fake_func2')
        self.assertEqual('fake_func2', getattr(result, 'name'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from karbor.common import constants
from karbor import exception
from karbor import resource
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protectable_plugin
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

INVALID_SHARE_STATUS = ['deleting', 'deleted', 'error', 'error_deleting',
                        'manage_error', 'unmanage_error', 'extending_error',
                        'shrinking_error', 'reverting_error']


class ShareProtectablePlugin(protectable_plugin.ProtectablePlugin):
    """Manila share protectable plugin"""

    _SUPPORT_RESOURCE_TYPE = constants.SHARE_RESOURCE_TYPE

    def _client(self, context):
        self._client_instance = ClientFactory.create_client(
            "manila",
            context)

        return self._client_instance

    def get_resource_type(self):
        return self._SUPPORT_RESOURCE_TYPE

    def get_parent_resource_types(self):
        return (constants.PROJECT_RESOURCE_TYPE, )

    def list_resources(self, context, parameters=None):
        try:
            shares = self._client(context).shares.list(detailed=True)
        except Exception as e:
            LOG.exception("List all summary shares from manila failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                      id=share.id, name=share.name)
                    for share in shares
                    if share.status not in INVALID_SHARE_STATUS]

    def show_resource(self, context, resource_id, parameters=None):
        try:
            share = self._client(context).shares.get(resource_id)
        except Exception as e:
            LOG.exception("Show a summary share from manila failed.")
            raise exception.ProtectableResourceNotFound(
                id=resource_id,
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            if share.status in INVALID_SHARE_STATUS:
                raise exception.ProtectableResourceInvalidStatus(
                    id=resource_id, type=self._SUPPORT_RESOURCE_TYPE,
                    status=share.status)
            return resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                     id=share.id, name=share.name)

    def get_dependent_resources(self, context, parent_resource):
        try:
            shares = self._client(context).shares.list()
        except Exception as e:
            LOG.exception("List all shares from manila failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                      id=share.id,
                                      name=share.name)
                    for share in shares
                    if share.project_id == parent_resource.id
                    and share.status not in INVALID_SHARE_STATUS]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_service import wsgi as base_wsgi

from karbor.api.openstack import ProjectMapper
from karbor.api.v1 import plans
from karbor.api.v1 import protectables
from karbor.api.v1 import providers
from karbor.api.v1 import restores
from karbor.api.v1 import scheduled_operations
from karbor.api.v1 import triggers


class APIRouter(base_wsgi.Router):
    @classmethod
    def factory(cls, global_conf, **local_conf):
        return cls(ProjectMapper())

    def __init__(self, mapper):
        plans_resources = plans.create_resource()
        restores_resources = restores.create_resource()
        protectables_resources = protectables.create_resource()
        providers_resources = providers.create_resource()
        trigger_resources = triggers.create_resource()
        scheduled_operation_resources = scheduled_operations.create_resource()

        mapper.resource("plan", "plans",
                        controller=plans_resources,
                        collection={},
                        member={'action': 'POST'})
        mapper.resource("restore", "restores",
                        controller=restores_resources,
                        collection={},
                        member={'action': 'POST'})
        mapper.resource("protectable", "protectables",
                        controller=protectables_resources,
                        collection={},
                        member={})
        mapper.connect("protectable",
                       "/{project_id}/protectables/"
                       "{protectable_type}/instances",
                       controller=protectables_resources,
                       action='instances_index',
                       conditions={"method": ['GET']})
        mapper.connect("protectable",
                       "/{project_id}/protectables/"
                       "{protectable_type}/instances/{protectable_id}",
                       controller=protectables_resources,
                       action='instances_show',
                       conditions={"method": ['GET']})
        mapper.resource("provider", "providers",
                        controller=providers_resources,
                        collection={},
                        member={})
        mapper.connect("provider",
                       "/{project_id}/providers/{provider_id}/checkpoints",
                       controller=providers_resources,
                       action='checkpoints_index',
                       conditions={"method": ['GET']})
        mapper.connect("provider",
                       "/{project_id}/providers/{provider_id}/checkpoints",
                       controller=providers_resources,
                       action='checkpoints_create',
                       conditions={"method": ['POST']})
        mapper.connect("provider",
                       "/{project_id}/providers/{provider_id}/checkpoints/"
                       "{checkpoint_id}",
                       controller=providers_resources,
                       action='checkpoints_show',
                       conditions={"method": ['GET']})
        mapper.connect("provider",
                       "/{project_id}/providers/{provider_id}/checkpoints/"
                       "{checkpoint_id}",
                       controller=providers_resources,
                       action='checkpoints_delete',
                       conditions={"method": ['DELETE']})
        mapper.resource("trigger", "triggers",
                        controller=trigger_resources,
                        collection={},
                        member={'action': 'POST'})
        mapper.resource("scheduled_operation", "scheduled_operations",
                        controller=scheduled_operation_resources,
                        collection={},
                        member={'action': 'POST'})
        super(APIRouter, self).__init__(mapper)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import timeutils

from karbor import exception
from karbor import objects
from karbor.tests.unit import objects as test_objects

NOW = timeutils.utcnow().replace(microsecond=0)

Log_ID = 0

Fake_Log = {
    'created_at': NOW,
    'deleted_at': None,
    'updated_at': NOW,
    'deleted': False,
    'id': Log_ID,
    'operation_id': '123',
    'expect_start_time': NOW,
    'triggered_time': NOW,
    'actual_start_time': NOW,
    'end_time': NOW,
    'state': 'in_progress',
    'extend_info': '',
}


class TestScheduledOperationLog(test_objects.BaseObjectsTestCase):
    def setUp(self):
        super(TestScheduledOperationLog, self).setUp()

        self.log_class = objects.ScheduledOperationLog
        self.db_log = Fake_Log

    @mock.patch('karbor.db.scheduled_operation_log_get')
    def test_get_by_id(self, log_get):
        log_get.return_value = self.db_log

        log = self.log_class.get_by_id(self.context, Log_ID)
        self._compare(self, self.db_log, log)
        log_get.assert_called_once_with(self.context, Log_ID)

    def test_get_by_no_existing_id(self):
        self.assertRaises(exception.ScheduledOperationLogNotFound,
                          self.log_class.get_by_id,
                          self.context, Log_ID)

    @mock.patch('karbor.db.scheduled_operation_log_create')
    def test_create(self, log_create):
        log_create.return_value = self.db_log

        log = self.log_class(context=self.context)
        log.create()
        self._compare(self, self.db_log, log)
        log_create.assert_called_once_with(self.context, {})

        self.assertRaises(exception.ObjectActionError,
                          log.create)

    @mock.patch('karbor.db.scheduled_operation_log_update')
    def test_save(self, log_update):
        log = self.log_class._from_db_object(self.context,
                                             self.log_class(),
                                             self.db_log)
        log.state = 'success'
        log.save()

        log_update.assert_called_once_with(self.context,
                                           log.id,
                                           {'state': 'success'})

    @mock.patch('karbor.db.scheduled_operation_log_delete')
    def test_destroy(self, log_delete):
        log = self.log_class._from_db_object(self.context,
                                             self.log_class(),
                                             self.db_log)
        log.destroy()
        log_delete.assert_called_once_with(self.context, log.id)


class TestScheduledOperationLogList(test_objects.BaseObjectsTestCase):

    def test_get_by_filters(self):
        log = self._create_operation_log('123')

        logs = objects.ScheduledOperationLogList.get_by_filters(
            self.context, {'state': ['in_progress']})
        self.assertEqual(1, len(logs.objects))
        log1 = logs.objects[0]
        self.assertEqual(log.id, log1.id)

    def _create_operation_log(self, operation_id):
        log_info = {
            'operation_id': operation_id,
            'state': 'in_progress',
        }
        log = objects.ScheduledOperationLog(self.context, **log_info)
        log.create()
        return log



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor.tests import base

from karbor import utils


class WalkClassHierarchyTestCase(base.TestCase):
    def test_walk_class_hierarchy(self):
        class A(object):
            pass

        class B(A):
            pass

        class C(A):
            pass

        class D(B):
            pass

        class E(A):
            pass

        class_pairs = zip((D, B, E),
                          utils.walk_class_hierarchy(A, encountered=[C]))
        for actual, expected in class_pairs:
            self.assertEqual(expected, actual)

        class_pairs = zip((D, B, C, E), utils.walk_class_hierarchy(A))
        for actual, expected in class_pairs:
            self.assertEqual(expected, actual)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from karbor.api.v1 import triggers as trigger_api
from karbor import context
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor.tests import base
from karbor.tests.unit.api import fakes


class FakeRemoteOperationApi(object):
    def create_trigger(self, context, trigger):
        if trigger.type not in ['time']:
            msg = (_("Invalid trigger type:%s") % trigger.type)
            raise exception.InvalidInput(msg)

        if trigger.properties['format'] not in ['crontab']:
            msg = (_("Invalid trigger time format type"))
            raise exception.InvalidInput(msg)

    def delete_trigger(self, context, trigger_id):
        pass

    def update_trigger(self, context, trigger):
        pass


class TriggerApiTest(base.TestCase):
    def setUp(self):
        super(TriggerApiTest, self).setUp()
        self.controller = trigger_api.TriggersController()
        self.controller.operationengine_api = FakeRemoteOperationApi()
        self.ctxt = context.RequestContext('demo', 'fakeproject',
                                           True)
        self.req = fakes.HTTPRequest.blank('/v1/triggers')
        self.default_create_trigger_param = {
            "name": "123",
            "type": "time",
            "properties": {
                "format": "crontab",
                "pattern": "* * * * *"
            },
        }

    def test_create_trigger_InvalidBody(self):
        self.assertRaises(exc.HTTPUnprocessableEntity,
                          self.controller.create,
                          self.req, {})

    def test_create_trigger_InvalidName(self):
        body = self._get_create_trigger_request_body()
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.req, body)

    def test_create_trigger_invalid_trigger_type(self):
        param = self.default_create_trigger_param.copy()
        param['type'] = "123"
        body = self._get_create_trigger_request_body(param)
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.req, body)

    def test_create_trigger_invalid_trigger_formt_type(self):
        param = self.default_create_trigger_param.copy()
        param['properties']['format'] = "123"
        body = self._get_create_trigger_request_body(param)
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.req, body)

    def test_create_trigger(self):
        name = 'every minutes'
        param = self.default_create_trigger_param.copy()
        param['name'] = name
        body = self._get_create_trigger_request_body(param)
        trigger = self.controller.create(self.req, body)
        self.assertEqual(name, trigger['trigger_info']['name'])

    def test_delete_trigger_binded_with_operation(self):
        trigger = self._create_one_trigger()
        trigger_id = trigger['trigger_info']['id']
        self._create_scheduled_operation(trigger_id)

        self.assertRaises(exc.HTTPFailedDependency,
                          self.controller.delete,
                          self.req,
                          trigger_id)

    def test_delete_trigger(self):
        trigger = self._create_one_trigger()
        self.controller.delete(self.req, trigger['trigger_info']['id'])
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show,
                          self.req,
                          trigger['trigger_info']['id'])

    def test_update_trigger(self):
        trigger = self._create_one_trigger()

        name = 'every minutes'
        param = self.default_create_trigger_param.copy()
        param['name'] = name
        param['properties']['window'] = 10
        body = self._get_create_trigger_request_body(param)
        trigger1 = self.controller.update(
            self.req, trigger['trigger_info']['id'], body)

        self.assertEqual(name, trigger1['trigger_info']['name'])
        self.assertEqual(10, int(
            trigger1['trigger_info']['properties']['window']))

    def test_show_trigger_not_exist(self):
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show,
                          self.req,
                          '2a9ce1f3-cc1a-4516-9435-0ebb13caa398')

    def test_show_trigger_invalid_id(self):
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.show,
                          self.req, 1)

    def test_show_trigger(self):
        trigger = self._create_one_trigger()
        trigger1 = self.controller.show(self.req,
                                        trigger['trigger_info']['id'])
        self.assertEqual(trigger['trigger_info']['id'],
                         trigger1['trigger_info']['id'])

    def test_list_trigger(self):
        trigger = self._create_one_trigger()
        triggers = self.controller.index(self.req)
        for item in triggers['triggers']:
            if item['id'] == trigger['trigger_info']['id']:
                self.assertTrue(1)

        self.assertFalse(0)

    def _create_one_trigger(self):
        param = self.default_create_trigger_param.copy()
        body = self._get_create_trigger_request_body(param)
        return self.controller.create(self.req, body)

    def _get_create_trigger_request_body(self, param={}):
        return {"trigger_info": param}

    def _create_scheduled_operation(self, trigger_id):
        operation_info = {
            "name": "123",
            "description": "123",
            "operation_type": "protect",
            'user_id': '123',
            "project_id": "123",
            "trigger_id": trigger_id,
            "operation_definition": {
                "plan_id": ""
            },
        }
        operation = objects.ScheduledOperation(self.ctxt, **operation_info)
        operation.create()
        return operation



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_versionedobjects import fields

from karbor import db
from karbor import exception
from karbor.i18n import _
from karbor.objects import base

CONF = cfg.CONF


@base.KarborObjectRegistry.register
class CheckpointRecord(base.KarborPersistentObject, base.KarborObject,
                       base.KarborObjectDictCompat,
                       base.KarborComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(),
        'project_id': fields.UUIDField(),
        'checkpoint_id': fields.UUIDField(),
        'checkpoint_status': fields.StringField(),
        'provider_id': fields.UUIDField(),
        'plan_id': fields.UUIDField(),
        'operation_id': fields.StringField(nullable=True),
        'create_by': fields.StringField(nullable=True),
        'extend_info': fields.StringField(nullable=True),
    }

    @staticmethod
    def _from_db_object(context, checkpoint_record, db_checkpoint_record):
        for name, field in checkpoint_record.fields.items():
            checkpoint_record[name] = db_checkpoint_record.get(name)

        checkpoint_record._context = context
        checkpoint_record.obj_reset_changes()
        return checkpoint_record

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        db_checkpoint_record = db.checkpoint_record_get(context, id)
        if db_checkpoint_record:
            return cls._from_db_object(context, cls(), db_checkpoint_record)

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        updates = self.karbor_obj_get_changes()
        db_checkpoint_record = db.checkpoint_record_create(self._context,
                                                           updates)
        self._from_db_object(self._context, self, db_checkpoint_record)

    @base.remotable
    def save(self):
        updates = self.karbor_obj_get_changes()
        if updates and self.id:
            db.checkpoint_record_update(self._context, self.id, updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        if self.id:
            db.checkpoint_record_destroy(self._context, self.id)


@base.KarborObjectRegistry.register
class CheckpointRecordList(base.ObjectListBase, base.KarborObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('CheckpointRecord'),
    }

    @base.remotable_classmethod
    def get_by_filters(cls, context, filters, limit=None,
                       marker=None, sort_keys=None, sort_dirs=None):

        checkpoint_record_list = db.checkpoint_record_get_all_by_filters_sort(
            context, filters, limit=limit, marker=marker,
            sort_keys=sort_keys, sort_dirs=sort_dirs)

        return base.obj_make_list(context,
                                  cls(context),
                                  CheckpointRecord,
                                  checkpoint_record_list)
